/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

package org.glite.ce.cream.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import org.glite.ce.creamapi.ws.cream2.types.BaseFaultType;
import org.glite.ce.creamapi.ws.cream2.types.Command;
import org.glite.ce.creamapi.ws.cream2.types.JobInfo;
import org.glite.ce.creamapi.ws.cream2.types.Lease;
import org.glite.ce.creamapi.ws.cream2.types.Status;

import org.glite.security.delegation.DelegationException;
import org.glite.security.delegation.Delegation;
import org.glite.security.delegation.GrDPX509Util;
import org.glite.security.delegation.GrDProxyGenerator;

import org.glite.ce.creamapi.ws.cream2.CREAMLocator;
import org.glite.ce.creamapi.ws.cream2.CREAMPort;
import org.glite.security.delegation.GrDProxyDlgorOptions;
import org.glite.security.delegation.DelegationServiceLocator;

public abstract class CmdLineTool {
    public static final int CURRENT_DELEGATION = 0;
    public static final int NEW_DELEGATION = 1;
    public static final int RENEW_DELEGATION = 2;

    /**
     * Creates a new CREAMPort object. Caller must supply the endPoint (in the
     * form <tt>https://<host>:<port></tt>, an (optional) delegationId, and
     * a boolean flag "delegate". This method accesses a configuration file
     * located in $home/.glite/dlgor.properties which <b>must</b> be present.
     * The file usually contains the following:
     * 
     * <pre>
     *     
     *      issuerCertFile=/tmp/x509up_u&lt;uid&gt;
     *      issuerKeyFile=/tmp/x509up_u&lt;uid&gt;
     *      issuerPass=
     *      delegationStorage=/tmp/dlgor_delegationStorage
     * </pre>
     * 
     * where <uid> is the Unix user ID.
     * 
     * @param endPoint
     *            The base URL of the CREAM Service. It usually must be in the
     *            form <tt>https://<host>:<port></tt>. The CREAM Service is
     *            assumed to be reachable through the URI:
     *            <tt><endPoint>/ce-cream/services/CREAM</tt>. The Delegation
     *            service is assumed to be reachable through the URI:
     *            <tt><endPoint>/ce-cream/services/CREAMDelegation</tt>.
     * @param delegationId
     *            The delegation ID to be used. This parameter is ignored is
     *            delegate is false; if delegationId is null, and delegate is
     *            true, then a new delegationId is automatically defined and
     *            used.
     * @param delegate
     *            If true, a proxy is delegated on the CREAM service. If false,
     *            then no delegation occurs. Proxy delegation is needed only if
     *            method CREAMPort::jobRegister() has to be called.
     * @return the CREAMPort object on which CREAM operations can be invoked.
     * @throws MalformedURLException
     *             if <tt><endPoint>/ce-cream/services/CREAM</tt> or
     *             <tt><endPoint>/ce-cream/services/CREAMDelegation</tt> are
     *             not valie URL.
     */
    static public CREAMPort getCREAM(URL endPoint, String delegationId, int delegationType) throws RuntimeException {
        if (delegationType > 0 && delegationId == null) {
            throw (new RuntimeException("the delegationId is null"));
        }

        String confFileName = System.getProperty("user.home") + "/.glite/dlgor.properties";
        if (endPoint.getProtocol().startsWith("https")) {
            try {
                GrDProxyDlgorOptions dlgorOpt = new GrDProxyDlgorOptions(confFileName);
                System.setProperty("axis.socketSecureFactory", "org.glite.security.trustmanager.axis.AXISSocketFactory");
                System.setProperty("gridProxyFile", dlgorOpt.getDlgorCertFile());
            } catch (Exception ex) {
                throw new RuntimeException(ex.getMessage());
            }
        }

        CREAMLocator creamLocator = new CREAMLocator();
        DelegationServiceLocator delegationLocator = new DelegationServiceLocator();

        if (delegationType > 0) {
            System.out.println("Using Delegation ID: " + delegationId);
            try {
                URL delegationPort = new URL(endPoint.toString() + "/ce-cream/services/gridsite-delegation");
                System.out.println("Delegation Port: " + delegationPort.toString());
                /*
                 * The delegation step is done in two phases. During the first
                 * phase, the user requests a proxy from CREAM (the pkcs10
                 * string). The pkcs10 string is associated with the
                 * delegationId specified by the user, which will be used in the
                 * following to identify the proxy. The pkcs10 string is the
                 * signed with the user's private key and sent back to CREAM
                 * with the delegation.putProxy() call.
                 */
                Delegation delegation = delegationLocator.getGridsiteDelegation(delegationPort);
                String pkcs10 = null;

                switch (delegationType) {
                case 1:
                    pkcs10 = delegation.getProxyReq(delegationId);
                    break;
                case 2:
                    pkcs10 = delegation.renewProxyReq(delegationId);
                    break;
                default:
                    throw (new Exception("wrong delegation type"));
                }

                String request = signRequest(pkcs10, delegationId);
                delegation.putProxy(delegationId, request);
            } catch (DelegationException e) {
                System.out.println("DelegationException: " + e.getMsg());
                System.exit(1);
            } catch (ServiceException e) {
                System.out.println("ServiceException: " + e);
                System.exit(1);
            } catch (RemoteException e) {
                System.out.println("RemoteException: " + e);
                System.exit(1);
            } catch (Exception e) {
                System.out.println("Exception: " + e);
                System.exit(1);
            }
        }
        URL creamPort = null;
        try {
            creamPort = new URL(endPoint.toString() + "/ce-cream/services/CREAM2");
        } catch (MalformedURLException e) {
            throw new RuntimeException("Malformed URL " + endPoint.toString() + "/ce-cream/services/CREAM2");
        }

        CREAMPort cream = null;
        try {
            cream = creamLocator.getCREAM2(creamPort);
        } catch (ServiceException e2) {
            System.out.println("ServiceException: " + e2);
            System.exit(1);
        }
        return cream;
    }

    static public Delegation getDelegation(URL endPoint) throws RuntimeException, ServiceException {
        String confFileName = System.getProperty("user.home") + "/.glite/dlgor.properties";
        if (endPoint.getProtocol().startsWith("https")) {
            try {
                GrDProxyDlgorOptions dlgorOpt = new GrDProxyDlgorOptions(confFileName);
                System.setProperty("axis.socketSecureFactory", "org.glite.security.trustmanager.axis.AXISSocketFactory");
                System.setProperty("gridProxyFile", dlgorOpt.getDlgorCertFile());
            } catch (Exception ex) {
                throw new RuntimeException(ex.getMessage());
            }
        }

        CREAMLocator creamLocator = new CREAMLocator();
        DelegationServiceLocator delegationLocator = new DelegationServiceLocator();

        URL epr = null;
        if (endPoint.getPath() == null || endPoint.getPath().length() == 0) {
            try {
                epr = new URL(endPoint.toString() + "/ce-cream/services/gridsite-delegation");
            } catch (MalformedURLException e) {
                throw new RuntimeException("Malformed URL " + endPoint.toString() + "/ce-cream/services/CREAM2");
            }
        } else {
            epr = endPoint;
        }
        /*
         * The delegation step is done in two phases. During the first phase,
         * the user requests a proxy from CREAM (the pkcs10 string). The pkcs10
         * string is associated with the delegationId specified by the user,
         * which will be used in the following to identify the proxy. The pkcs10
         * string is the signed with the user's private key and sent back to
         * CREAM with the delegation.putProxy() call.
         */
        return delegationLocator.getGridsiteDelegation(epr);
    }


    protected static String signRequest(String certReq, String delegationID) throws IOException {
        String strX509CertChain = null;
        String confFileName = System.getProperty("user.home") + "/.glite/dlgor.properties";
        GrDProxyDlgorOptions dlgorOpt = new GrDProxyDlgorOptions(confFileName);

        try {
            GrDProxyGenerator proxyGenerator = new GrDProxyGenerator();

            byte[] x509Cert = proxyGenerator.x509MakeProxyCert(certReq.getBytes(), GrDPX509Util.getFilesBytes(new File(dlgorOpt.getDlgorCertFile())), "null");

            strX509CertChain = new String(x509Cert);
        } catch (Exception e) {
            e.printStackTrace();
        }

        return strX509CertChain;
    }

    
    static public CREAMPort getCREAM(URL endPoint) throws RuntimeException, ServiceException {
        String confFileName = System.getProperty("user.home") + "/.glite/dlgor.properties";
        if (endPoint.getProtocol().startsWith("https")) {
            try {
                GrDProxyDlgorOptions dlgorOpt = new GrDProxyDlgorOptions(confFileName);
                System.setProperty("axis.socketSecureFactory", "org.glite.security.trustmanager.axis.AXISSocketFactory");
                System.setProperty("gridProxyFile", dlgorOpt.getDlgorCertFile());
            } catch (Exception ex) {
                throw new RuntimeException(ex.getMessage());
            }
        }

        CREAMLocator creamLocator = new CREAMLocator();

        URL epr = null;
        if (endPoint.getPath() == null || endPoint.getPath().length() == 0) {
            try {
                epr = new URL(endPoint.toString() + "/ce-cream/services/CREAM2");
            } catch (MalformedURLException e) {
                throw new RuntimeException("Malformed URL " + endPoint.toString() + "/ce-cream/services/CREAM2");
            }
        } else {
            epr = endPoint;
        }

        return creamLocator.getCREAM2(epr);
    }
    
    /**
     * Reads a file and returns a String representation of its content.
     * 
     * @param filename
     *            the complete path of the file to read.
     * @return the content of file filename.
     * @throws Exception
     *             if the file does not exists or cannot be read.
     */
    protected final static String readFile(String filename) throws IOException {
        // String res = new String();
        // String line = null;
        // BufferedReader in = new BufferedReader(new FileReader(filename));
        // do {
        // line = in.readLine();
        // if (line != null)
        // res += "\n" + line;
        // } while (line != null);
        // in.close();
        // return res;

        String jdl = "";
        FileReader in = null;
        try {
            in = new FileReader(filename);
        } catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        char[] buffer = new char[1024];
        int n = 1;

        while (n > 0) {
            try {
                n = in.read(buffer, 0, buffer.length);
            } catch (IOException e) {
                e.printStackTrace();
            }

            if (n > 0) {
                jdl += new String(buffer, 0, n);
            }
        }
        try {
            in.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return jdl;
    }
}
