/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

package org.glite.ce.cream.client;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.rpc.ServiceException;

import org.glite.ce.creamapi.ws.cream2.types.BaseFaultType;
import org.glite.ce.creamapi.ws.cream2.types.JobDescription;
import org.glite.ce.creamapi.ws.cream2.types.JobId;
import org.glite.ce.creamapi.ws.cream2.types.JobRegisterResult;
import org.glite.ce.creamapi.ws.cream2.types.Property;
import org.glite.ce.creamapi.ws.cream2.CREAMLocator;
import org.glite.ce.creamapi.ws.cream2.CREAMPort;
import org.glite.security.delegation.Delegation;
import org.glite.security.delegation.DelegationException;
import org.glite.security.delegation.GrDPX509Util;
import org.glite.security.delegation.GrDProxyDlgorOptions;
import org.glite.security.delegation.GrDProxyGenerator;

public class JobRegister extends JobCommand {

    public JobRegister(String[] args, List<String> options) throws RuntimeException {
        super(args, options);
    }
    
    protected void execute() {
        String delegProxy = null;
        
        if(isDelegate()) {
            try {
                Delegation delegation = getCreamDelegation();
                String pkcs10 = delegation.getProxyReq(getDelegationId());
                delegProxy = signRequest(pkcs10, getDelegationId());
                //delegation.putProxy(getDelegationId(), request);
            } catch (ServiceException e) {
                System.out.println("ServiceException: " + e.getMessage());
                System.exit(0);
            } catch (DelegationException e) {
                System.out.println("DelegationException: " + e.getMsg());
                System.exit(0);
            } catch (RemoteException e) {
                System.out.println("RemoteException: " + e.getMessage());
                System.exit(0);
            } catch (IOException e) {
                System.out.println("IOException: " + e.getMessage());
                System.exit(0);
            }        
        }
        
        try {
            String[] jdlFileArray = getJdlFileArray();
            if(jdlFileArray == null) {
                System.out.println("no jdl specified!");
                return;
            }

            JobDescription[] jobDescr = new JobDescription[jdlFileArray.length];            
            for(int i=0; i<jdlFileArray.length; i++) {
                String jdl = readFile(jdlFileArray[i]);

                jobDescr[i] = new JobDescription();
                jobDescr[i].setJobDescriptionId(""+i);
                jobDescr[i].setJDL(jdl);
                jobDescr[i].setAutoStart(isAutostart());
                jobDescr[i].setDelegationId(getDelegationId());
                jobDescr[i].setDelegationProxy(delegProxy);
                jobDescr[i].setLeaseId(getLeaseId());                
            }
            
            JobRegisterResult[] result = getCream().jobRegister(jobDescr);

            for (int i = 0; i < result.length; i++) {
                System.out.println("job descrption id: " + result[i].getJobDescriptionId());

                if (result[i].getJobId() != null) {
                    JobId jobId = result[i].getJobId();
                    System.out.println("jobId: " + jobId.getId());
                    System.out.println("cream url: " + jobId.getCreamURL());

                    Property[] property = jobId.getProperty();

                    if (property != null) {
                        for (int x = 0; x < property.length; x++) {
                            System.out.println("property: name=" + property[x].getName() + " value=\"" + property[x].getValue());
                        }
                    }
                } else {
                    System.out.println("fault returned: ");

                    if(result[i].getDelegationProxyFault() != null) {
                        printFault(result[i].getDelegationProxyFault());
                    } else if(result[i].getDelegationIdMismatchFault() != null) {
                        printFault(result[i].getDelegationIdMismatchFault());
                    } else if(result[i].getGenericFault() != null) {
                        printFault(result[i].getGenericFault()); 
                    } else if(result[i].getLeaseIdMismatchFault() != null) {
                        printFault(result[i].getLeaseIdMismatchFault());
                    }
                }
            }
        } catch (BaseFaultType e) {
            printFault(e);
        } catch (RemoteException e) {
            System.out.println("RemoteException: " + e.getMessage());
        } catch (ServiceException e) {
            System.out.println("ServiceException: " + e.getMessage());
        } catch (IOException e) {
            System.out.println("IOException: " + e.getMessage());
        }
        
    }
    
    public static void main(String[] args) {
        List<String> options = new ArrayList<String>();
        options.add(EPR);
        options.add(DELEGATION_ID);
        options.add(LEASE_ID);
        options.add(AUTOSTART);
        options.add(DELEGATE);
        options.add(JDL_FILE);
        
        new JobRegister(args, options);
    }
}
