/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
package org.glite.ce.cream.cmdmanagement;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.cream.cmdmanagement.db.CommandDBManager;
import org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueDBManager;
import org.glite.ce.cream.configuration.ServiceConfig;
import org.glite.ce.cream.configuration.CREAMConfigConstants.ServiceAttribute;
import org.glite.ce.creamapi.cmdmanagement.Command;
import org.glite.ce.creamapi.cmdmanagement.CommandException;
import org.glite.ce.creamapi.cmdmanagement.CommandExecutor;
import org.glite.ce.creamapi.cmdmanagement.CommandManagementException;
import org.glite.ce.creamapi.cmdmanagement.CommandManagerInterface;
import org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueException;
import org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface;

public class CommandManager implements CommandManagerInterface {
    private final static Logger logger = Logger.getLogger(CommandManager.class.getName());
    private static CommandManager commandManager = null;
    
    public static CommandManager getInstance() throws CommandManagementException {
        if (commandManager == null) {
            commandManager = new CommandManager();
            commandManager.init();
        }

        return commandManager;
    }

    private Hashtable<String, Hashtable<String, CommandExecutor>> cmdExecutorCategory;
    private Hashtable<String, CommandQueueInterface> cmdQueueTable;
    
    private CommandDBManager commandDBManager;
    
    private CommandManager() {
        cmdExecutorCategory = new Hashtable<String, Hashtable<String, CommandExecutor>>(0);
        cmdQueueTable = new Hashtable<String, CommandQueueInterface>(0);
    }

    public void addCommandExecutor(CommandExecutor cmdExec) throws CommandManagementException, IllegalArgumentException {
        if (cmdExec == null) {
            throw new IllegalArgumentException("The CommandExecutor is null!");
        }

        if (cmdExec.getCategory() == null) {
            throw new CommandManagementException("CommandExecutor's category not defined!");
        }

        if (cmdExec.getName() == null) {
            throw new CommandManagementException("CommandExecutor's name not defined!");
        }

        Hashtable<String, CommandExecutor> executorTable = cmdExecutorCategory.get(cmdExec.getCategory());

        if (executorTable == null) {
            executorTable = new Hashtable<String, CommandExecutor>(0);
            cmdExecutorCategory.put(cmdExec.getCategory(), executorTable);
        }

        if (executorTable.containsKey(cmdExec.getName())) {
            logger.debug("The CommandExecutor \"" + cmdExec.getName() + "\" already exists");
            throw new CommandManagementException("The CommandExecutor \"" + cmdExec.getName() + "\" already exists");
        } else {
            try {
                cmdExec.setCommandManager(this);
                cmdExec.initExecutor();

                ServiceConfig serviceConf = ServiceConfig.getConfiguration();

                if (serviceConf == null) {
                    throw new CommandManagementException("Unable to get the service configuration!");
                }
                
                CommandQueueInterface queue = null;
                
                if(cmdQueueTable.containsKey(cmdExec.getCategory())) {
                    queue = cmdQueueTable.get(cmdExec.getCategory());
                } else {
                    try {
                        queue = new CommandQueueDBManager(cmdExec.getCategory(), CommandManagerInterface.COMMAND_DATASOURCE_NAME);  //using the default datasource
                        String shared = serviceConf.getStringParameter(ServiceAttribute.CREAM_COMMAND_QUEUE_SHARED);
                        if(shared != null && shared.equalsIgnoreCase("true")) {
                            queue.setFetchCommandTimeout(60000);
                        }
                        queue.open();
                    } catch (CommandQueueException e) {
                        throw new CommandManagementException(e.getMessage());
                    }
                    cmdQueueTable.put(cmdExec.getCategory(), queue);
                }
                     
                cmdExec.setCommandQueue(queue);
                
                int cl = 5;
                if(serviceConf != null) {
                    cl = serviceConf.getIntParameter(ServiceAttribute.CREAM_CONCURRENCY_LEVEL, 5);
                }
                
                cmdExec.setCommandWorkerPoolSize(cl);

                executorTable.put(cmdExec.getName(), cmdExec);

                logger.debug("New CommandExecutor \"" + cmdExec.getName() + "\" added");
            } catch (CommandException e) {
                logger.error(cmdExec.getName() + " initialization failed: " + e.getMessage());
                throw new CommandManagementException(cmdExec.getName() + " initialization failed: " + e.getMessage());
            }
        }
    }

    public void addCommandExecutor(CommandExecutor[] cmdExec) throws CommandManagementException, IllegalArgumentException {
        if (cmdExec != null) {
            for (int i = 0; i < cmdExec.length; i++) {
                addCommandExecutor(cmdExec[i]);
            }
        } else {
            throw new IllegalArgumentException("The CommandExecutor array is null!");
        }
    }

    public void addCommands(ArrayList cmd) throws CommandException, IllegalArgumentException {
        if (cmd != null) {
            for (int i = 0; i < cmd.size(); i++) {
                insertCommand((Command) cmd.get(i));
            }
        }
    }

    public boolean checkCommandExecutorSupport(String name, String category) {
        if (name == null || category == null) {
            return false;
        }

        if (cmdExecutorCategory.containsKey(category)) {
            Hashtable<String, CommandExecutor> executorTable = cmdExecutorCategory.get(category);
            return executorTable.containsKey(name);
        }

        return false;
    }

    public void deleteCommand(String cmdId) throws CommandException, IllegalArgumentException {
        if(cmdId == null) {
            throw new IllegalArgumentException("cmdId not defined!");
        }
        commandDBManager.deleteCommand(cmdId);
    }

    public void destroy() throws CommandManagementException {
        Collection<CommandQueueInterface> queueList = cmdQueueTable.values();
        if (queueList != null) {
            Iterator<CommandQueueInterface> iter = queueList.iterator();
            CommandQueueInterface queue = null;

            while (iter.hasNext()) {
                queue = iter.next();
                queue.close();
            }
        }

        if (!cmdExecutorCategory.isEmpty()) {
            Iterator<Hashtable<String, CommandExecutor>> iter = cmdExecutorCategory.values().iterator();
            Hashtable<String, CommandExecutor> executorList = null;

            while (iter.hasNext()) {
                executorList = iter.next();
                CommandExecutor executor = null;

                if (!executorList.isEmpty()) {
                    Iterator<CommandExecutor> iter2 = executorList.values().iterator();

                    while (iter2.hasNext()) {
                        executor = iter2.next();
                        executor.destroy();
                    }
                }
            }
        }
    }

    public Command getCommand(String cmdId) throws CommandException, IllegalArgumentException {
        if(cmdId == null) {
            throw new IllegalArgumentException("cmdId not defined!");
        }

        return commandDBManager.getCommand(cmdId);
    }

    public List<String> getCommandCategory() {
        List<String> categoryList = new ArrayList<String>(0);
        if (!cmdExecutorCategory.isEmpty()) {
            categoryList.addAll(cmdExecutorCategory.keySet());
        }
        return categoryList;
    }

    public List<CommandExecutor> getCommandExecutor() {
        List<CommandExecutor> commandExecutorList = new ArrayList<CommandExecutor>(0);
        
        if (!cmdExecutorCategory.isEmpty()) {            
            Hashtable cmdExecutortable = null;
            Iterator<Hashtable<String, CommandExecutor>> iter = cmdExecutorCategory.values().iterator();

            while (iter.hasNext()) {
                cmdExecutortable = iter.next();
                commandExecutorList.addAll(cmdExecutortable.values());
            }
        }
        return commandExecutorList;
    }

    public CommandExecutor getCommandExecutor(String name, String category) throws CommandManagementException, IllegalArgumentException {
        if (name == null) {
            throw new IllegalArgumentException("name not specified!");
        }

        if (category != null) {
            if (cmdExecutorCategory.containsKey(category)) {
                Hashtable<String, CommandExecutor> cmdExecutortable = cmdExecutorCategory.get(category);
                if (cmdExecutortable.containsKey(name)) {
                    return cmdExecutortable.get(name);
                }
                throw new CommandManagementException("CommandExecutor \"" + name + "\" not found!");
            }
            throw new CommandManagementException("Category \"" + category + "\" not found!");
        } else {
            Hashtable<String, CommandExecutor> cmdExecutortable = null;
            Iterator<Hashtable<String, CommandExecutor>> iter = cmdExecutorCategory.values().iterator();

            while (iter.hasNext()) {
                cmdExecutortable = iter.next();

                if (cmdExecutortable.containsKey(name)) {
                    return cmdExecutortable.get(name);
                }
            }
            throw new CommandManagementException("CommandExecutor \"" + name + "\" not found!");
        }
    }

    public void init() throws CommandManagementException {
        try {
            commandDBManager = new CommandDBManager(CommandManagerInterface.COMMAND_DATASOURCE_NAME);
        } catch (DatabaseException e) {
            throw new CommandManagementException(e.getMessage());
        }  
    }

    public void insertCommand(Command cmd) throws CommandException, IllegalArgumentException {
        if(cmd == null) {
            throw new IllegalArgumentException("command not defined!");
        }

        if (cmd.getCategory() == null) {
            throw new CommandException("Command category not defined!");
        }

        Hashtable<String, CommandExecutor> executorTable = cmdExecutorCategory.get(cmd.getCategory());

        if (executorTable == null) {
            throw new CommandException("Command category \"" + cmd.getCategory() + "\" not found!");
        }

        CommandExecutor executor = null;

        if (cmd.getCommandExecutorName() == null) {
            Collection<CommandExecutor> cmdExec = executorTable.values();
            ArrayList<CommandExecutor> list = new ArrayList<CommandExecutor>(cmdExec);
            int index = 0;
            while(executor == null && index < list.size()) {
                executor = list.get(index++);
            }

            if (executor == null) {
                throw new CommandException("None right executor can handle the command \"" + cmd.getName() + "\"");
            }
            
        } else {
            logger.debug("cmd.getCommandExecutorName() = " + cmd.getCommandExecutorName());
            executor = executorTable.get(cmd.getCommandExecutorName());

            if (executor == null) {
                throw new CommandException("CommandExecutor \"" + cmd.getCommandExecutorName() + "\" not found!");
            }
        }

        logger.debug("Selected the \"" + executor.getName() + "\" executor for the execution of the cmd \"" + cmd.getName() + "\"");

        cmd.setCommandExecutorName(executor.getName());
        if (executor.checkCommandSupport(cmd.getName())) {
            logger.debug("Setting command status to SCHEDULED (" + Command.SCHEDULED + ")");
            cmd.setStatus(Command.SCHEDULED);

            if (cmd.isAsynchronous()) {
            	logger.debug("Command is Asynchronous");
                try {
                    logger.debug("Inserting the command into the database ...");
                    
                    long startTime = 0L;

                    if (logger.isDebugEnabled()) {
                        startTime = Calendar.getInstance().getTimeInMillis();
                    }

                    commandDBManager.insertCommand(cmd);

                    if (logger.isDebugEnabled()) {
                        logger.debug("The command has been inserted! (" + (Calendar.getInstance().getTimeInMillis()-startTime) + " msec)");
                    }
                } catch (Throwable e) {
                    logger.error("cannot insert the command " + cmd.getName() + " into the database: " + e.getMessage());
                    
                    throw new CommandException("cannot insert the command " + cmd.getName() + " into the database: " + e.getMessage());
                }
                
                try {
                    logger.debug("Enqueueing the command " + cmd.getName());

                    long startTime = 0L;

                    if (logger.isDebugEnabled()) {
                        startTime = Calendar.getInstance().getTimeInMillis();
                    }

                    executor.getCommandQueue().enqueueCommand(cmd);

                    if (logger.isDebugEnabled()) {
                        logger.debug("The command has been enqueued! (" + (Calendar.getInstance().getTimeInMillis()-startTime) + " msec)");
                    }
                } catch (Throwable e) {
                    logger.error("cannot enqueue the command " + cmd.getName() + ": " + e.getMessage());
//                    logger.error(e);
//                    cmd.setStatus(Command.ERROR);
//                    cmd.setFailureReason(e.getMessage());
//                    logger.debug("Updating command ...");
//                    commandDBManager.updateCommand(cmd);
//                    logger.debug("Command has been updated.");
                    logger.debug("Deleting the command " + cmd.getName());
                    commandDBManager.deleteCommand(cmd.getId());
                    logger.debug("The command has been deleted.");
                    
                    throw new CommandException("cannot enqueue the command " + cmd.getName() + ": " + e.getMessage());
                }
            } else {
            	logger.debug("Command is Not Asynchronous");
                try {
                    cmd.setStatus(Command.PROCESSING);
                    // commandDBManager.updateCommand(cmd);
                    logger.debug("CommandManager insertCommand: executing the command " + cmd.getName() + " by the executor " + executor.getName());

                    executor.execute(cmd);
                    logger.debug("CommandManager insertCommand: " + cmd.getName() + " done");

                    cmd.setStatus(Command.SUCCESSFULL);
                   // commandDBManager.updateCommand(cmd);
                  //  commandDBManager.deleteCommand(cmd.getId());
                } catch (Exception e) {
                	logger.error("CommandManager insertCommand: " + cmd.getName() + " error " + e.getMessage());
                   // cmd.setStatus(Command.ERROR);
                   // cmd.setFailureReason(e.getMessage());
                   // commandDBManager.updateCommand(cmd);
                    
                    throw new CommandException(e.getMessage());
                }
//                if(!cmd.isSuccessfull()) {
//                    logger.debug("Command execution error: " + cmd.getFailureReason());
//                }
            }
        } else {
            throw new CommandException("Command \"" + cmd.getName() + "\" not supported by the category \"" + cmd.getCategory() + "\" !");
        }
        logger.debug("END insertCommand");
    }


    public void insertCommands(Command[] cmd) throws CommandException {
        if(cmd == null) {
            throw new IllegalArgumentException("command not defined!");
        }
        
        if (cmd != null) {
            for (int i = 0; i < cmd.length; i++) {
                insertCommand(cmd[i]);
            }
        }
    }

    public void removeCommandExecutor(CommandExecutor cmdExec) throws CommandManagementException, IllegalArgumentException {
        if(cmdExec == null) {
            throw new IllegalArgumentException("command executor not defined!");
        }

        removeCommandExecutor(cmdExec.getName(), cmdExec.getCategory());
    }

    public void removeCommandExecutor(CommandExecutor[] cmdExec) throws CommandManagementException, IllegalArgumentException {
        if(cmdExec == null) {
            throw new IllegalArgumentException("command executor not defined!");
        }

        for (int i = 0; i < cmdExec.length; i++) {
            if (cmdExec[i] != null) {
                removeCommandExecutor(cmdExec[i].getName(), cmdExec[i].getCategory());
            }
        }
    }

    public void removeCommandExecutor(String name, String category) throws CommandManagementException, IllegalArgumentException {
        if (category == null) {
            throw new IllegalArgumentException("CommandExecutor's category not defined!");
        }

        if (name == null) {
            throw new IllegalArgumentException("CommandExecutor's name not defined!");
        }

        if (cmdExecutorCategory.containsKey(category)) {
            Hashtable<String, CommandExecutor> executorTable = cmdExecutorCategory.get(category);

            CommandExecutor cmdExec = executorTable.remove(name);
            if (cmdExec != null) {
                cmdExec.destroy();
                logger.debug("CommandExecutor \"" + name + "\" removed");
                
                if(executorTable.size() == 0) {
                    CommandQueueInterface queue = cmdQueueTable.remove(category);
                    queue.close();
                    logger.debug("CommandQueue \"" + category + "\" closed");
                    queue.destroy();
                    logger.debug("CommandQueue \"" + category + "\" destroyed");
                }
            } else {
                throw new CommandManagementException("CommandExecutor \"" + name + "\" not found!");
            }
        } else {
            throw new CommandManagementException("Category \"" + category + "\" not found!");
        }
    }
    
    public void updateCommand(Command cmd) throws CommandException, IllegalArgumentException {
        if (cmd == null) {
            throw new IllegalArgumentException("command not defined!");
        }
        commandDBManager.updateCommand(cmd, false);
    }

    public void updateCommand(Command cmd, boolean lightUpdate) throws CommandException, IllegalArgumentException {
        if (cmd == null) {
            throw new IllegalArgumentException("command not defined!");
        }
        commandDBManager.updateCommand(cmd, lightUpdate);
    }
}
