/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.common.db.DatasourceManager;
import org.glite.ce.creamapi.cmdmanagement.Command;
import org.glite.ce.creamapi.cmdmanagement.CommandException;

/**
 * Class managing commands in the database.
 * 
 */
public class CommandDBManager { // implements CommandManagerInterface {

    /** The logger */
    private static final Logger logger = Logger.getLogger(CommandDBManager.class);

    /** The datasource name */
    private String datasourceName = null;

    /**
     * DAO for managing command table.
     */
    private CommandTable commandTable = null;

    /**
     * DAO for managing command parameter table.
     */
    private CommandParameterTable commandParameterTable = null;

    /**
     * Default constructor.
     * 
     * @throws DatabaseException
     */
    public CommandDBManager(String datasourceName) throws DatabaseException, IllegalArgumentException {
        if (datasourceName == null) {
            throw new IllegalArgumentException("datasourceName not specified!");
        }

        // Init datasource name
        this.datasourceName = datasourceName;

        // Init DAOs
        commandTable = new CommandTable();
        commandParameterTable = new CommandParameterTable();
    }

    /**
     * Returns the connection to the database.
     * 
     * @return The connection to the database.
     * @throws DatabaseException
     */
    private Connection getConnection() throws CommandException {
        Connection connection = null;
        try {
            connection = DatasourceManager.getConnection(datasourceName);
        } catch (DatabaseException ex) {
            throw new CommandException(ex.getMessage());
        }

        if (connection == null) {
            logger.error("cannot get the database connection");
            throw new CommandException("cannot get the database connection");
        }

        return connection;
    }

    /**
     * @throws DatabaseException
     * @see org.glite.ce.creamapi.cmdmanagement.CommandManagerInterface#deleteCommand(java.lang.String)
     */
    public void deleteCommand(String commandId) throws CommandException, IllegalArgumentException {
        logger.debug("deleteCommand begin (cmdId=" + commandId + ")");

        long startTime = 0L;

        if (logger.isDebugEnabled()) {
            startTime = Calendar.getInstance().getTimeInMillis();
        }

        Connection connection = getConnection();

        try {
            commandTable.executeDelete(commandId, connection);

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandException(sqle2.getMessage());
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("deleteCommand end (cmdId=" + commandId + ") (" + (Calendar.getInstance().getTimeInMillis() - startTime) + " msec)");
        }
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.CommandManagerInterface#getCommand(java.lang.String)
     */
    public Command getCommand(String commandId) throws CommandException, IllegalArgumentException {
        logger.debug("getCommand begin (cmdId=" + commandId + ")");

        long startTime = 0L;

        if (logger.isDebugEnabled()) {
            startTime = Calendar.getInstance().getTimeInMillis();
        }

        Connection connection = getConnection();
        Command cmd = null;

        try {
            cmd = commandTable.executeSelect(commandId, connection);
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandException(sqle2.getMessage());
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("getCommand end (cmdId=" + commandId + ") (" + (Calendar.getInstance().getTimeInMillis() - startTime) + " msec)");
        }

        return cmd;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.CommandManagerInterface#insertCommand(org.glite.ce.creamapi.cmdmanagement.Command)
     */
    public void insertCommand(Command cmd) throws CommandException, IllegalArgumentException {
        logger.debug("insertCommand begin (cmdId=" + cmd.getId() + ")");

        long startTime = 0L;

        if (logger.isDebugEnabled()) {
            startTime = Calendar.getInstance().getTimeInMillis();
        }

        Connection connection = getConnection();

        try {
            commandTable.executeInsert(cmd, connection);

            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandException(sqle2.getMessage());
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("insertCommand end (cmdId=" + cmd.getId() + ") (" + (Calendar.getInstance().getTimeInMillis() - startTime) + " msec)");
        }
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.CommandManagerInterface#updateCommand(org.glite.ce.creamapi.cmdmanagement.Command)
     */
    public void updateCommand(Command cmd, boolean lightUpdate) throws CommandException, IllegalArgumentException {
        logger.debug("updateCommand begin (cmdId=" + cmd.getId() + ")");

        long startTime = 0L;

        if (logger.isDebugEnabled()) {
            startTime = Calendar.getInstance().getTimeInMillis();
        }

        Connection connection = getConnection();

        try {
            // Update
            commandTable.executeUpdate(cmd, connection);

            if (!lightUpdate) {
                // Delete command parameters if present
                List<CommandParameter> parameters = commandParameterTable.findByCommandId(Long.parseLong(cmd.getId()), connection);

                for (CommandParameter parameter : parameters) {
                    commandParameterTable.executeDelete(parameter.getId(), connection);
                    logger.debug("Deleted parameter with name=" + parameter.getName() + ", value=" + parameter.getValue());
                }

                // Insert command parameters if present
                List<CommandParameter> newParameters = getParameterCommands(cmd);

                for (CommandParameter parameter : newParameters) {
                    parameter.setCommandId(Long.parseLong(cmd.getId()));

                    commandParameterTable.executeInsert(parameter, connection);
                    logger.debug("Inserted parameter with name=" + parameter.getName() + ", value=" + parameter.getValue());
                }
            }
            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandException(sqle2.getMessage());
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("updateCommand end (cmdId=" + cmd.getId() + ") (" + (Calendar.getInstance().getTimeInMillis() - startTime) + " msec)");
        }
    }

    /**
     * Returns the list of command parameters.
     * 
     * @param cmd
     *            The command.
     * @return The list of command parameters.
     */
    private List<CommandParameter> getParameterCommands(Command cmd) {
        ArrayList<CommandParameter> parameters = new ArrayList<CommandParameter>(0);

        if (cmd != null) {
            Set<String> keys = cmd.getParameterKeySet();

            for (String key : keys) {
                List<String> valueList = cmd.getParameterMultivalue(key);

                for (String value : valueList) {
                    parameters.add(new CommandParameter(key, value));
                }
            }
        }

        return parameters;
    }
}
