/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.AbstractTable;
import org.glite.ce.common.db.Query;

/**
 * Class representing the Data Access Object (DAO) to the command table in the database.
 *
 */
public class CommandParameterTable extends AbstractTable implements CommandParameterTableInterface  {

    /** The logger */
    private final static Logger logger = Logger.getLogger(CommandParameterTable.class);

    /** The table name */
    public final static String NAME_TABLE = "command_parameter";

    // The table columns
    public final static String ID_FIELD = "id";
    public final static String NAME_FIELD = "name";
    public final static String VALUE_FIELD = "value";
    public final static String COMMAN_ID_FIELD = "commandId";

    /** Sequence name */
    public static final String SEQUENCE_NAME = "command_parameter_id_seq";

    /** Label for find by command id prepared statement */
    public static final String FIND_BY_COMMAND_ID_STATEMENT = "FIND_BY_COMMAND_ID_STATEMENT";

    /** The find by command id query */
    protected Query findByCommandIdQuery = null;

    
    public CommandParameterTable() {
        insertQuery = makeInsertQuery();
        deleteQuery = makeDeleteQuery();
        selectQuery = makeSelectQuery();
        findByCommandIdQuery = makeFindByCommandIdQuery();
    }
    
    
    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandParameterTableInterface#executeInsert(org.glite.ce.cream.cmdmanagement.db.CommandParameter, java.sql.Connection)
     */
    public void executeInsert(CommandParameter parameter, Connection connection) throws SQLException, IllegalArgumentException {
        if (parameter == null) {
            throw new IllegalArgumentException("The object is null");
        }
        // Check the non nullable fields
        if (parameter.getCommandId() <= 0) {
            throw new IllegalArgumentException("The command object has an invalid commandId=" + parameter.getCommandId());
        }

        logger.debug("executeInsert begin");

        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            pstmt = getPreparedStatement(insertQuery, connection);
            pstmt.setLong(1, parameter.getCommandId());
            pstmt.setString(2, parameter.getName());
            pstmt.setString(3, parameter.getValue());

            int rowCount = pstmt.executeUpdate();

            // Get autogenerated keys
            rset = pstmt.getGeneratedKeys();

            if (rset.next()) {
                parameter.setId(rset.getLong(1));
            } else {
                throw new SQLException("Problem in retrieving autogenerated keys");
            }

            logger.debug("inserted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }

        }
        logger.debug("executeInsert end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandParameterTableInterface#executeDelete(long, java.sql.Connection)
     */
    public void executeDelete(long parameterId, Connection connection) throws SQLException, IllegalArgumentException {
        if (parameterId < 0) {
            throw new IllegalArgumentException("The id = " + parameterId + " is invalid");
        }

        logger.debug("executeDelete begin");
        PreparedStatement pstmt = null;
        try {
            pstmt = getPreparedStatement(deleteQuery, connection);
            pstmt.setLong(1, parameterId);

            int rowCount = pstmt.executeUpdate();
            logger.debug("deleted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeDelete end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandParameterTableInterface#executeSelect(long, java.sql.Connection)
     */
    public CommandParameter executeSelect(long parameterId, Connection connection) throws SQLException, IllegalArgumentException {
        if (parameterId < 0) {
            throw new IllegalArgumentException("The id = " + parameterId + " is invalid");
        }

        logger.debug("executeSelect begin");

        CommandParameter parameter = null;
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            pstmt = getPreparedStatement(selectQuery, connection);
            pstmt.setLong(1, parameterId);

            rset = pstmt.executeQuery();
            if (rset != null) {
                rset.next();
                parameter = new CommandParameter(rset.getString(3), rset.getString(4));
                parameter.setId(rset.getLong(1));
                parameter.setCommandId(rset.getLong(2));
                rset.close();
            }
            logger.debug("executeSelect end");
            return parameter;
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandParameterTableInterface#findByCommandId(long, java.sql.Connection)
     */
    public List<CommandParameter> findByCommandId(long commandId, Connection connection) throws SQLException, IllegalArgumentException {
        if (commandId < 0) {
            throw new IllegalArgumentException("The id = " + commandId + " is invalid");
        }

        logger.debug("findByCommandId begin");
        ArrayList<CommandParameter> parameters = new ArrayList<CommandParameter>(0);
        CommandParameter parameter = null;
        PreparedStatement pstmt = null;
        ResultSet rset = null;

        try {
            pstmt = getPreparedStatement(findByCommandIdQuery, connection);
            pstmt.setLong(1, commandId);

            rset = pstmt.executeQuery();

            if (rset != null) {
                while (rset.next()) {
                    parameter = new CommandParameter(rset.getString(3), rset.getString(4));
                    parameter.setId(rset.getLong(1));
                    parameter.setCommandId(rset.getLong(2));
                    parameters.add(parameter);
                }
            }

            logger.debug("findByCommandId end");
            return parameters;
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
    }

    /**
     * Returns the query for inserting
     * @return The query for inserting.
     */
    protected Query makeInsertQuery() {
        StringBuffer query = new StringBuffer();
        query.append("insert into ");
        query.append(NAME_TABLE);
        query.append(" ( ");
        query.append(COMMAN_ID_FIELD + ", ");
        query.append(NAME_FIELD + ", ");
        query.append(VALUE_FIELD);
        query.append(") values(?, ?, ?)");

        Query insertQuery = new Query();
        insertQuery.setName(Query.INSERT_STATEMENT);
        insertQuery.setStatement(query.toString());

        return insertQuery;
    }

    /**
     * Returns the query for deleting.
     * @return The query for deleting.
     */
    protected Query makeDeleteQuery() {
        StringBuffer query = new StringBuffer();
        query.append("delete from ");
        query.append(NAME_TABLE);
        query.append(" where ");
        query.append(ID_FIELD);
        query.append(" = ");
        query.append(" ? ");

        Query deleteQuery = new Query();
        deleteQuery.setName(Query.DELETE_STATEMENT);
        deleteQuery.setStatement(query.toString());

        return deleteQuery;
    }

    /**
     * Returns the query for selecting.
     * @return The query for selecting.
     */
    protected Query makeSelectQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select ");
        query.append(ID_FIELD + ", ");
        query.append(COMMAN_ID_FIELD + ", ");
        query.append(NAME_FIELD + ", ");
        query.append(VALUE_FIELD);
        query.append(" from ");
        query.append(NAME_TABLE);
        query.append(" where ");
        query.append(ID_FIELD);
        query.append(" = ?");

        Query selectQuery = new Query();
        selectQuery.setName(Query.SELECT_STATEMENT);
        selectQuery.setStatement(query.toString());

        return selectQuery;
    }

    /**
     * Returns the query for finding by commandId.
     * @return The query for finding by commandId.
     */
    protected Query makeFindByCommandIdQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select ");
        query.append(ID_FIELD + ", ");
        query.append(COMMAN_ID_FIELD + ", ");
        query.append(NAME_FIELD + ", ");
        query.append(VALUE_FIELD);
        query.append(" from ");
        query.append(NAME_TABLE);
        query.append(" where ");
        query.append(COMMAN_ID_FIELD);
        query.append(" = ?");

        Query findByCommandIdQuery = new Query();
        findByCommandIdQuery.setName(FIND_BY_COMMAND_ID_STATEMENT);
        findByCommandIdQuery.setStatement(query.toString());

        return findByCommandIdQuery;
    }
}
