/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.AbstractTable;
import org.glite.ce.common.db.Query;
import org.glite.ce.creamapi.cmdmanagement.Command;

/**
 * Class representing the Data Access Object (DAO) to the command table in the
 * database.
 * 
 */
public class CommandTable extends AbstractTable implements CommandTableInterface {

    /** The logger */
    private final static Logger logger = Logger.getLogger(CommandTable.class);

    /** The table name */
    public final static String NAME_TABLE = "command";

    // The table columns
    public final static String ID_FIELD = "id";
    public final static String NAME_FIELD = "name";
    public final static String CATEGORY_FIELD = "category";
    public final static String DESCRIPTION_FIELD = "description";
    public final static String STATUS_TYPE_FIELD = "statusType";
    public final static String FAILURE_REASON_FIELD = "failureReason";
    public final static String CMD_EXECUTOR_NAME_FIELD = "cmdExecutorName";
    public final static String USER_ID_FIELD = "userId";
    public final static String START_SCHEDULING_TIME_FIELD = "startSchedulingTime";
    public final static String START_PROCESSING_TIME_FIELD = "startProcessingTime";
    public final static String EXECUTION_COMPLETED_TIME_FIELD = "executionCompletedTime";
    public final static String CREATION_TIME_FIELD = "creationTime";

    /** Sequence name */
    public static final String SEQUENCE_NAME = "command_id_seq";

    private CommandParameterTable commandParameterTable = null;

    public CommandTable() {
        insertQuery = makeInsertQuery();
        deleteQuery = makeDeleteQuery();
        selectQuery = makeSelectQuery();

        commandParameterTable = new CommandParameterTable();
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandTableInterface#executeInsert(org.glite.ce.creamapi.cmdmanagement.Command,
     *      java.sql.Connection)
     */
    public void executeInsert(Command cmd, Connection connection) throws SQLException, IllegalArgumentException {
        if (cmd == null) {
            throw new IllegalArgumentException("Command object not specified!");
        }
        // Check the non nullable fields
        if (cmd.getUserId() == null) {
            throw new IllegalArgumentException("The command object has a null userId");
        }
        if (cmd.getCreationTime() == null) {
            throw new IllegalArgumentException("The command object has a null creation time");
        }

        logger.debug("executeInsert begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;

        long commandId = -1;

        try {
            pstmt = getPreparedStatement(insertQuery, connection);
            pstmt.setString(1, cmd.getName());
            pstmt.setString(2, cmd.getCategory());
            pstmt.setString(3, cmd.getDescription());
            pstmt.setInt(4, cmd.getStatus());
            pstmt.setString(5, cmd.getFailureReason());
            pstmt.setString(6, cmd.getCommandExecutorName());
            pstmt.setString(7, cmd.getUserId());

            Timestamp startSchedulingTime = null;
            if (cmd.getStartSchedulingTime() != null && cmd.getStartSchedulingTime() != null) {
                startSchedulingTime = new Timestamp(cmd.getStartSchedulingTime().getTimeInMillis());
            }
            pstmt.setTimestamp(8, startSchedulingTime);

            Timestamp startProcessingTime = null;
            if (cmd.getStartProcessingTime() != null && cmd.getStartProcessingTime() != null) {
                startProcessingTime = new Timestamp(cmd.getStartProcessingTime().getTimeInMillis());
            }
            pstmt.setTimestamp(9, startProcessingTime);

            Timestamp executionCompletedTime = null;
            if (cmd.getExecutionCompletedTime() != null && cmd.getExecutionCompletedTime() != null) {
                executionCompletedTime = new Timestamp(cmd.getExecutionCompletedTime().getTimeInMillis());
            }
            pstmt.setTimestamp(10, executionCompletedTime);

            Timestamp creationTime = new Timestamp(cmd.getCreationTime().getTimeInMillis());
            pstmt.setTimestamp(11, creationTime);
            logger.debug("insert statetement = " + pstmt.toString());

            int rowCount = pstmt.executeUpdate();

            // Get autogenerated keys
            rset = pstmt.getGeneratedKeys();

            if (rset.next()) {
                commandId = rset.getLong(1);
                cmd.setId(Long.toString(commandId));
            } else {
                throw new SQLException("Problem in retrieving autogenerated keys");
            }

            logger.debug("inserted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }

        if (commandId > -1) {
            Set<String> keys = cmd.getParameterKeySet();
            CommandParameter parameter = null;

            for (String key : keys) {
                List<String> valueList = cmd.getParameterMultivalue(key);

                for (String value : valueList) {
                    parameter = new CommandParameter(key, value);
                    parameter.setCommandId(commandId);

                    commandParameterTable.executeInsert(parameter, connection);
                    logger.debug("Inserted parameter with name=" + parameter.getName() + ", value=" + parameter.getValue());
                }
            }

            logger.debug("executeInsert end");
        }
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandTableInterface#executeDelete(java.lang.String,
     *      java.sql.Connection)
     */
    public void executeDelete(String commandId, Connection connection) throws SQLException, IllegalArgumentException {
        if (commandId == null) {
            throw new IllegalArgumentException("Command id not specified!");
        }
        if (Long.parseLong(commandId) < 0) {
            throw new IllegalArgumentException("The id = " + commandId + " is invalid");
        }

        logger.debug("executeDelete begin " + commandId);
        PreparedStatement pstmt = null;
        try {
            pstmt = getPreparedStatement(deleteQuery, connection);
            pstmt.setLong(1, Long.parseLong(commandId));

            int rowCount = pstmt.executeUpdate();
            logger.debug("deleted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeDelete end " + commandId);
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandTableInterface#executeUpdate(org.glite.ce.creamapi.cmdmanagement.Command,
     *      java.sql.Connection)
     */
    public void executeUpdate(Command cmd, Connection connection) throws SQLException, IllegalArgumentException {
        if (cmd == null) {
            throw new IllegalArgumentException("Command object not specified!");
        }
        if (cmd.getId() == null || Integer.parseInt(cmd.getId()) < 0) {
            throw new IllegalArgumentException("The id = " + cmd.getId() + " is invalid");
        }

        logger.debug("executeUpdate begin");
        PreparedStatement pstmt = null;

        try {
            Query query = null;
            Calendar time = null;
            Timestamp timestamp = null;
            int status = cmd.getStatus();

            switch (status) {
            case Command.SCHEDULED:
                time = cmd.getStartSchedulingTime();
                break;
            case Command.RESCHEDULED:
                query = getUpdateTimeQuery(START_SCHEDULING_TIME_FIELD);
                break;
            case Command.PROCESSING:
                query = getUpdateTimeQuery(START_PROCESSING_TIME_FIELD);
                time = cmd.getStartProcessingTime();
                break;
            case Command.ERROR:
            case Command.REMOVED:
            case Command.SUCCESSFULL:
                query = getUpdateTimeQuery(EXECUTION_COMPLETED_TIME_FIELD);
                time = cmd.getExecutionCompletedTime();
            }

            pstmt = getPreparedStatement(query, connection);
            pstmt.setInt(1, cmd.getStatus());
            pstmt.setString(2, cmd.getFailureReason());
            pstmt.setString(3, cmd.getDescription());

            if (time != null) {
                timestamp = new Timestamp(time.getTimeInMillis());
            }

            pstmt.setTimestamp(4, timestamp);
            pstmt.setLong(5, Long.parseLong(cmd.getId()));

            logger.debug("update statetement = " + pstmt.toString());
            int rowCount = pstmt.executeUpdate();
            logger.debug("updated " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeUpdate end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.db.CommandTableInterface#executeSelect(java.lang.String,
     *      java.sql.Connection)
     */
    public Command executeSelect(String commandId, Connection connection) throws SQLException, IllegalArgumentException {
        if (commandId == null) {
            throw new IllegalArgumentException("Command id not specified!");
        }
        if (Long.parseLong(commandId) < 0) {
            throw new IllegalArgumentException("The id = " + commandId + " is invalid");
        }

        logger.debug("executeSelect begin");
        Command cmd = null;
        PreparedStatement pstmt = null;
        ResultSet rset = null;

        try {
            pstmt = getPreparedStatement(selectQuery, connection);
            pstmt.setLong(1, Long.parseLong(commandId));

            rset = pstmt.executeQuery();

            if (rset != null) {
                // Create multivalue map for the command parameters
                Map<String, List<String>> map = new HashMap<String, List<String>>(0);

                while (rset.next()) {
                    long cmdId = rset.getInt(1);

                    if (cmd == null) {
                        cmd = new Command(rset.getString(2));
                        cmd.setId(commandId);
                        cmd.setCategory(rset.getString(3));
                        cmd.setDescription(rset.getString(4));
                        cmd.setStatus(rset.getInt(5));
                        cmd.setFailureReason(rset.getString(6));
                        cmd.setCommandExecutorName(rset.getString(7));
                        cmd.setUserId(rset.getString(8));

                        Timestamp timestamp = rset.getTimestamp(9);
                        Calendar calendar = null;
                        if (timestamp != null) {
                            calendar = Calendar.getInstance();
                            calendar.setTimeInMillis(timestamp.getTime());
                            cmd.setStartSchedulingTime(calendar);
                        }
                        cmd.setStartSchedulingTime(calendar);

                        timestamp = rset.getTimestamp(10);
                        calendar = null;
                        if (timestamp != null) {
                            calendar = Calendar.getInstance();
                            calendar.setTimeInMillis(timestamp.getTime());
                            cmd.setStartProcessingTime(calendar);
                        }
                        cmd.setStartSchedulingTime(calendar);

                        timestamp = rset.getTimestamp(11);
                        calendar = null;
                        if (timestamp != null) {
                            calendar = Calendar.getInstance();
                            calendar.setTimeInMillis(timestamp.getTime());
                            cmd.setStartSchedulingTime(calendar);
                        }
                        cmd.setExecutionCompletedTime(calendar);

                        timestamp = rset.getTimestamp(12);
                        calendar = null;
                        if (timestamp != null) {
                            calendar = Calendar.getInstance();
                            calendar.setTimeInMillis(timestamp.getTime());
                            cmd.setStartSchedulingTime(calendar);
                        }
                        cmd.setCreationTime(calendar);
                    }

                    long parameterId = rset.getLong(13);

                    if (parameterId > 0) {
                        String parameterName = rset.getString(14);
                        String parameterValue = rset.getString(15);

                        List<String> valueList = map.get(parameterName);
                        if (valueList == null) {
                            map.put(parameterName, valueList = new ArrayList<String>(0));
                        }
                        valueList.add(parameterValue);
                    }
                }

                // Iterate over the map to set command parameters
                for (String key : map.keySet()) {
                    List<String> valueList = map.get(key);

                    if (valueList != null) {
                        if (valueList.size() == 1) {
                            cmd.addParameter(key, valueList.get(0));
                        } else {
                            cmd.addParameter(key, valueList);
                        }
                    }
                }
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeSelect end");
        return cmd;
    }

    /**
     * Returns the query for inserting
     * 
     * @return The query for inserting.
     */
    protected Query makeInsertQuery() {
        StringBuffer query = new StringBuffer();
        query.append("insert into ");
        query.append(NAME_TABLE);
        query.append(" ( ");
        query.append(NAME_FIELD + ", ");
        query.append(CATEGORY_FIELD + ", ");
        query.append(DESCRIPTION_FIELD + ", ");
        query.append(STATUS_TYPE_FIELD + ", ");
        query.append(FAILURE_REASON_FIELD + ", ");
        query.append(CMD_EXECUTOR_NAME_FIELD + ", ");
        query.append(USER_ID_FIELD + ", ");
        query.append(START_SCHEDULING_TIME_FIELD + ", ");
        query.append(START_PROCESSING_TIME_FIELD + ", ");
        query.append(EXECUTION_COMPLETED_TIME_FIELD + ", ");
        query.append(CREATION_TIME_FIELD);
        query.append(") values(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

        Query insertQuery = new Query();
        insertQuery.setName(Query.INSERT_STATEMENT);
        insertQuery.setStatement(query.toString());

        return insertQuery;
    }

    /**
     * Returns the query for deleting.
     * 
     * @return The query for deleting.
     */
    protected Query makeDeleteQuery() {
        StringBuffer query = new StringBuffer();
        query.append("delete from ");
        query.append(NAME_TABLE);
        query.append(" where ");
        query.append(ID_FIELD);
        query.append(" = ?");

        Query deleteQuery = new Query();
        deleteQuery.setName(Query.DELETE_STATEMENT);
        deleteQuery.setStatement(query.toString());

        return deleteQuery;
    }

    /**
     * Returns the query string for updating.
     * 
     * @param timeField
     *            The time field to be updated.
     * @return The query string for updating.
     */
    protected Query getUpdateTimeQuery(String timeField) {
        StringBuffer query = new StringBuffer();
        query.append("update ");
        query.append(NAME_TABLE);
        query.append(" set ");
        query.append(STATUS_TYPE_FIELD);
        query.append(" = ?, ");
        query.append(FAILURE_REASON_FIELD);
        query.append(" = ?, ");
        query.append(DESCRIPTION_FIELD);
        query.append(" = ?, ");
        query.append(timeField);
        query.append(" = ? where ");
        query.append(ID_FIELD);
        query.append(" = ?");

        Query updateQuery = new Query();
        updateQuery.setName(Query.UPDATE_STATEMENT + timeField);
        updateQuery.setStatement(query.toString());

        return updateQuery;
    }

    /**
     * Returns the query for selecting.
     * 
     * @return The query for selecting.
     */
    protected Query makeSelectQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select ");
        query.append(NAME_TABLE).append(".").append(ID_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(NAME_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(CATEGORY_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(DESCRIPTION_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(STATUS_TYPE_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(FAILURE_REASON_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(CMD_EXECUTOR_NAME_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(USER_ID_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(START_SCHEDULING_TIME_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(START_PROCESSING_TIME_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(EXECUTION_COMPLETED_TIME_FIELD).append(", ");
        query.append(NAME_TABLE).append(".").append(CREATION_TIME_FIELD).append(", ");
        query.append("command_parameter").append(".").append("id").append(", ");
        query.append("command_parameter").append(".").append("name").append(", ");
        query.append("command_parameter").append(".").append("value");
        query.append(" from ");
        query.append(NAME_TABLE);
        query.append(", command_parameter where ");
        query.append(NAME_TABLE).append(".").append(ID_FIELD);
        query.append(" = ? AND command_parameter.commandId=").append(NAME_TABLE).append(".").append(ID_FIELD);

        Query selectQuery = new Query();
        selectQuery.setName(Query.SELECT_STATEMENT);
        selectQuery.setStatement(query.toString());

        return selectQuery;
    }
}
