/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.queue.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.common.db.DatasourceManager;
import org.glite.ce.commonj.utils.BooleanLock;
//import org.glite.ce.cream.cmdmanagement.db.CommandParameterTable;
import org.glite.ce.cream.cmdmanagement.db.CommandTable;
import org.glite.ce.creamapi.cmdmanagement.Command;
import org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueException;
import org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface;


/**
 * Class managing a command queue in the database.
 * 
 */
public class CommandQueueDBManager implements CommandQueueInterface {
    /** The logger */
    private final static Logger logger = Logger.getLogger(CommandQueueDBManager.class);

    /** The datasource name */
    private String datasourceName = null;

    /** The command queue name */
    private String queueName;

    /**
     * DAO for managing command table.
     */
    private CommandTable commandTable = null;

    /**
     * DAO for managing command queue table.
     */
    private CommandQueueTable commandQueueTable = null;

    /**
     * DAO for managing command parameter table.
     */
 //   private CommandParameterTable commandParameterTable = null;

    /**
     * DAO for managing command queue list table.
     */
    private CommandQueueListTable commandQueueListTable = null;
    private boolean isOpen = false;
    private long msTimeout = 0L;
    private BlockingQueue<Command> queue;
    private final int maxQueueSize = 500; 
    private final BooleanLock fillUpQueueLock = new BooleanLock(true);
    private FillUpQueueThread fillUpQueueThread = null;

    /**
     * Constructor.
     * 
     * @param queueName
     *            The name of the queue, which corresponds to the same one of
     *            the associated table.
     * @param datasourceName
     *            The datasource name.
     * @throws CommandQueueException
     */
    public CommandQueueDBManager(String queueName, String datasourceName) throws CommandQueueException, IllegalArgumentException {
        if(datasourceName == null) {
            throw new IllegalArgumentException("datasourceName not specified!");
        }

        // Init datasource name
        this.datasourceName = datasourceName;

        if (queueName == null) {
            logger.error("queueName not defined!");
            throw new IllegalArgumentException("queueName not specified!");
        }

        this.queueName = queueName;

        // Init DAOs
        commandTable = new CommandTable();
        commandQueueTable = new CommandQueueTable(queueName);
//        commandParameterTable = new CommandParameterTable();
        commandQueueListTable = new CommandQueueListTable();

        queue = new LinkedBlockingQueue<Command>(maxQueueSize);

        // Inserts the queue table in the db if not present
        insertQueue(queueName);

        // Reset queue
        recoverQueue();
    }

    /**
     * Returns the connection to the database.
     * 
     * @return The connection to the database.
     * @throws DatabaseException
     */
    private Connection getConnection() throws CommandQueueException {
        Connection connection = null;
        try {
            connection = DatasourceManager.getConnection(datasourceName);
        } catch (DatabaseException ex) {
            throw new CommandQueueException(ex.getMessage());
        }

        if (connection == null) {
            logger.error("cannot get the database connection");
            throw new CommandQueueException("cannot get the database connection");
        }

        return connection;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#getName()
     */
    public String getName() {
        return queueName;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#dequeueCommand(java.lang.String)
     */
    public void dequeueCommand(String commandId) throws CommandQueueException, IllegalArgumentException {
        logger.debug("dequeueCommand begin");

        Connection connection = getConnection();

        try {
            // Delete
            commandQueueTable.executeDelete(commandId, connection);

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("dequeueCommand end");
    }

    public void setFetchCommandTimeout(long msTimeout) {
        this.msTimeout = msTimeout < 0L ? 60000 : msTimeout;
    }

    public Command fetchCommand() throws CommandQueueException, IllegalArgumentException {
        try {
            return queue.take();
        } catch (InterruptedException e) {
            return null;
        }
    }

    private class FillUpQueueThread extends Thread {
        public FillUpQueueThread() {
            super("FillUpQueueThread");
            setDaemon(true);
        }

        public void run() {
            while (isOpen) {
                if(queue.size() >= maxQueueSize) {
                    try {
                        Thread.sleep(1000L);
                    } catch (InterruptedException e) {}

                    continue;
                }

                long totalResult = 0L;
                Hashtable<String, ArrayList<Command>> cmdTable = new Hashtable<String, ArrayList<Command>>(0);
                Connection connection = null;

                try {
                    connection = getConnection();
                    int commandToBeEnqueued = 0;

                    for (int i = Command.HIGH_PRIORITY; i >= Command.LOW_PRIORITY && (maxQueueSize > (queue.size() + commandToBeEnqueued)); i--) {
                        // Find the first unscheduled command id from the command queue
                        List<CommandQueueItem> result = commandQueueTable.executeFindCommandInQueue(i, maxQueueSize - queue.size() - commandToBeEnqueued, connection);
                        totalResult += result.size();

                        Command cmd = null;

                        for (CommandQueueItem item : result) {                            
                            // There is at least one command to be extracted
                            logger.debug("There is a command to be extracted with commandId=" + item.getCommandId());
                            
                            cmd = commandTable.executeSelect(Long.toString(item.getCommandId()), connection);
                            cmd.setCommandGroupId(item.getCommandGroupId());
                            cmd.setAsynchronous(true);
                            cmd.setPriorityLevel(item.getPriorityLevel());
                            cmd.setExecutionMode(item.getExecutionMode());

                            ArrayList<Command> cmdList = null;

                            if (cmdTable.containsKey(cmd.getCommandGroupId())) {
                                cmdList = cmdTable.get(cmd.getCommandGroupId());
                            } else {
                                cmdList = new ArrayList<Command>(0);
                                cmdTable.put(cmd.getCommandGroupId(), cmdList);
                            }

                            boolean found = false;

                            for (Command c : cmdList) {
                                if(c.getExecutionMode().equals(Command.ExecutionModeValues.SERIAL)) {
                                    found = true;
                                    break;
                                }
                            }

                            if (!found) {
                                item.setScheduled(true);

                                // Update flag
                                commandQueueTable.executeUpdate(item, connection);
                                logger.debug("update executed: isScheduled set to true for commandId=" + item.getCommandId());

                                connection.commit();
                                cmdList.add(cmd);
                                commandToBeEnqueued++;
                            }
                        }
                    }

                    String key = null;
                    ArrayList<Command> cmdList = null;

                    for (Enumeration<String> e = cmdTable.keys(); e.hasMoreElements(); ) {
                        queue.addAll(cmdTable.get(e.nextElement()));
                    }
                } catch (CommandQueueException e) {
                    logger.error("FillUpQueueThread error: " + e.getMessage());
                } catch (IllegalStateException e) {
                    logger.error("FillUpQueueThread error: " + e.getMessage());
                } catch (SQLException sqle) {           
                    try {
                        logger.error("FillUpQueueThread: operation failed: " + sqle.getMessage());
                        if (connection != null) {
                            connection.rollback();
                            logger.error("Rollback performed");
                        }
                    } catch (SQLException sqle1) {
                        logger.error("Rollback failed: " + sqle1.getMessage());
                        // throw new CommandQueueException(sqle1.getMessage());
                    }
                    //  throw new CommandQueueException(sqle.getMessage());
                } finally {
                    if (connection != null) {
                        try {
                            connection.close();
                        } catch (SQLException sqle2) {
                            logger.error("Problem in closing connection: " + sqle2.getMessage());
                            //    throw new CommandQueueException(sqle2.getMessage());
                        }
                    }
                }
                if(totalResult == 0L) {
                    try {
                        fillUpQueueLock.waitToSetFalse(getQueueSize() == 0L? msTimeout: 5000L);
                    } catch (Throwable e) {
                        logger.error("FillUpQueueThread error: " + e.getMessage());
                    } 
                }
            }
        }
    }

    public int getQueueSize() throws CommandQueueException {
        logger.debug("getQueueSize begin");

        Connection connection = getConnection();
        int queueSize = 0;

        try {
            queueSize = commandQueueTable.executeGetQueueSize(connection);
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("getQueueSize end"); 

        return queueSize;
    }

    public boolean isOpen() {
        return isOpen;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#open()
     */
    public void open() {
        isOpen = true;
        fillUpQueueThread = new FillUpQueueThread();
        fillUpQueueThread.start();
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#close()
     */
    public void close() {
        isOpen = false;
        fillUpQueueLock.setValue(true);
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#destroy()
     */
    public void destroy() {
        close();
    }

    /**
     * Returns the command corresponding to the first command id in queue to be
     * extracted if any are present, null otherwise.
     * 
     * @return The command corresponding to the first command id in queue to be
     *         extracted if any are present, null otherwise.
     * @throws CommandQueueException
     */
    public Command retrieveFirstCommandInQueue() throws CommandQueueException {
        logger.debug("retrieveFirstCommandInQueue begin");

        Command cmd = null;
        Connection connection = getConnection();

        try {
            for (int i = Command.HIGH_PRIORITY; i >= Command.LOW_PRIORITY; i--) {
                // Find the first unscheduled command id from the command queue
                CommandQueueItem item = commandQueueTable.executeFindFirstCommandInQueue(i, connection);

                if (item != null) {
                    // There is at least one command to be extracted
                    logger.debug("There is a command to be extracted with commandId=" + item.getCommandId());

                    item.setScheduled(true);

                    // Update flag
                    commandQueueTable.executeUpdate(item, connection);
                    logger.debug("update executed: isScheduled set to true for commandId=" + item.getCommandId());

                    connection.commit();

                    cmd = commandTable.executeSelect(Long.toString(item.getCommandId()), connection);
                    cmd.setCommandGroupId(item.getCommandGroupId());
                    cmd.setAsynchronous(true);
                    cmd.setPriorityLevel(item.getPriorityLevel());
                    cmd.setExecutionMode(item.getExecutionMode());

                    break;
                }
            }
        } catch (SQLException sqle) {
            logger.error(sqle.getMessage(), sqle);

            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("retrieveFirstCommandInQueue end");
        return cmd;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#recoverQueue()
     */
    public void recoverQueue() throws CommandQueueException, IllegalArgumentException {
        logger.debug("recoverQueue begin");

        Connection connection = getConnection();

        try {
            // Get the list of scheduled commands in queue
            List<CommandQueueItem> commands = commandQueueTable.executeFindScheduledCommands(connection);

            for (CommandQueueItem item : commands) {
                // Update the item retrieved (isScheduled -> false)
                item.setScheduled(false);
                item.setPriorityLevel(Command.HIGH_PRIORITY);

                commandQueueTable.executeUpdate(item, connection);

                Command cmd = commandTable.executeSelect(Long.toString(item.getCommandId()), connection);
                if (cmd != null) {
                    cmd.setStatus(Command.RESCHEDULED);
                    cmd.setDescription("command rescheduled by CREAM");
                    commandTable.executeUpdate(cmd, connection);
                }
            }

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("recoverQueue end");
    }

    /**
     * Deletes a given queue name.
     * 
     * @param queueName
     *            The queue name to be deleted.
     * @throws CommandQueueException
     * @throws IllegalArgumentException
     */
    public void deleteQueue(String queueName) throws CommandQueueException, IllegalArgumentException {
        logger.debug("deleteQueue begin");

        Connection connection = getConnection();

        try {
            // Drop table
            commandQueueListTable.executeTableDrop(queueName, connection);

            // Delete queue name
            commandQueueListTable.executeDelete(queueName, connection);

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandQueueException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("deleteQueue end");
    }

    /**
     * Inserts a given queue name.
     * 
     * @param queueName
     *            The queue name to be inserted.
     * @throws CommandQueueException
     */
    public void insertQueue(String queueName) throws CommandQueueException {
        logger.debug("insertQueue begin");

        Connection connection = getConnection();

        try {
            if (!isQueueNamePresent(queueName)) {
                // Generate the table if it does not exist
                commandQueueListTable.executeTableCreation(queueName, connection);

                // Insert the queue name in the database queue list
                commandQueueListTable.executeInsert(queueName, connection);
            } else {
                logger.debug("Queue name is already existing");
            }

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new CommandQueueException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("insertQueue end");
    }

    /**
     * Checks if a queue name exists in the list of queues.
     * 
     * @queueName The queue name.
     * @return Returns true if the queue exists, false otherwise.
     * @throws CommandQueueException
     */
    public boolean isQueueNamePresent(String queueName) throws CommandQueueException {
        boolean result = false;
        logger.debug("isQueueNamePresent begin");

        Connection connection = getConnection();

        try {
            // Find the first unscheduled command id from the command queue
            result = commandQueueListTable.executeIsQueueNamePresent(queueName, connection);
            logger.debug("Executed executeIsQueueNamePresent result=" + result);

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("isQueueNamePresent end");
        return result;
    }

    /**
     * Returns the list of queue names registered at a given time. If no
     * matching occurs the returned list is empty.
     * 
     * @return Returns the list of queue names registered at a given time. If no
     *         matching occurs the returned list is empty.
     * @throws CommandQueueException
     */
    public List<String> findQueueNames() throws CommandQueueException {
        logger.debug("findQueueNames begin");

        Connection connection = getConnection();
        List<String> names = new ArrayList<String>(0);

        try {
            // Find the first unscheduled command id from the command queue
            names = commandQueueListTable.executeFindQueueNames(connection);
            logger.debug("Executed executeFindQueueNames");

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        logger.debug("findQueueNames end");
        return names;
    }

    /**
     * @see org.glite.ce.creamapi.cmdmanagement.queue.CommandQueueInterface#enqueueCommand(org.glite.ce.creamapi.cmdmanagement.Command,
     *      int)
     */
    public void enqueueCommand(Command cmd) throws CommandQueueException, IllegalArgumentException {
        logger.debug("enqueueCommand begin");

        Connection connection = getConnection();

        try {
            long commandId = cmd.getId() != null ? Long.parseLong(cmd.getId()) : -1;

            CommandQueueItem item = new CommandQueueItem();
            item.setCommandId(commandId);
            item.setCommandGroupId(cmd.getCommandGroupId());
            item.setExecutionMode(cmd.getExecutionMode());
            item.setPriorityLevel(cmd.getPriorityLevel());

            // Insert into command queue table
            commandQueueTable.executeInsert(item, connection);

            // Commit
            connection.commit();

            fillUpQueueLock.setValue(true);
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new CommandQueueException(sqle1.getMessage());
            }
            throw new CommandQueueException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new CommandQueueException(sqle2.getMessage());
                }
            }
        }
        
        logger.debug("enqueueCommand end");
    }
}
