/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.queue.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.AbstractTable;
import org.glite.ce.common.db.Query;

/**
 * Class representing the Data Access Object to a command queue table in the
 * database.
 * 
 */
public class CommandQueueListTable extends AbstractTable implements CommandQueueListTableInterface {

    /** The logger */
    private final static Logger logger = Logger.getLogger(CommandQueueListTable.class);

    /** The table name: it changes according to the command queue name */
    public final static String NAME_TABLE = "command_queue_list";

    // The table columns
    public final static String NAME = "name";

    /** Query name to check if a queue name exists in the queue list */
    public static final String IS_QUEUE_NAME_PRESENT = "IS_QUEUE_NAME_PRESENT";

    /** Query name to find queue names */
    public static final String FIND_QUEUE_NAMES_QUERY = "FIND_QUEUE_NAMES_QUERY";

    /** Query name to create a new queue table */
    public static final String TABLE_CREATION_QUERY = "TABLE_CREATION_QUERY";

    /** Query name to drop a new queue table */
    public static final String TABLE_DROP_QUERY = "TABLE_DROP_QUERY";

    /** Query to check if a queue name exists in the queue list */
    protected Query isQueueNamePresentQuery = null;

    /** Query to find queue names */
    protected Query findQueueNamesQuery = null;
        

    public CommandQueueListTable() {
        insertQuery = makeInsertQuery();
        deleteQuery = makeDeleteQuery();
        isQueueNamePresentQuery = makeIsQueueNamePresentQuery();
        findQueueNamesQuery = makeFindQueueNamesQuery(); 
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueListTableInterface#executeInsert(java.lang.String,
     *      java.sql.Connection)
     */
    public void executeInsert(String queueName, Connection connection) throws SQLException, IllegalArgumentException {
        if (queueName == null) {
            throw new IllegalArgumentException("The queue name is null");
        }

        logger.debug("executeInsert begin");

        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            pstmt = getPreparedStatement(insertQuery, connection);
            pstmt.setString(1, queueName);
            
            int rowCount = pstmt.executeUpdate();
            logger.debug("inserted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeInsert end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeDelete(java.lang.String,
     *      java.sql.Connection)
     */
    public void executeDelete(String queueName, Connection connection) throws SQLException, IllegalArgumentException {
        if (queueName == null) {
            throw new IllegalArgumentException("The queue name is null");
        }

        logger.debug("executeDelete begin");
        PreparedStatement pstmt = null;

        try {
            pstmt = getPreparedStatement(deleteQuery, connection);
            pstmt.setString(1, queueName);
            
            int rowCount = pstmt.executeUpdate();
            logger.debug("deleted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeDelete end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueListTableInterface#executeIsQueueNamePresent(java.lang.String,
     *      java.sql.Connection)
     */
    public boolean executeIsQueueNamePresent(String queueName, Connection connection) throws SQLException, IllegalArgumentException {
        boolean result = false;
        logger.debug("executeIsQueueNamePresent begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            // Check if a queue name is registered
            pstmt = getPreparedStatement(isQueueNamePresentQuery, connection);
            pstmt.setString(1, queueName);
            
            rset = pstmt.executeQuery();
            if (rset != null) {
                rset.next();
                if (rset.getRow() > 0) {
                    long count = rset.getLong(1);
                    if (count > 0) {
                        result = true;
                    }
                }
            }
            logger.debug("check result = " + result);
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeIsQueueNamePresent end");
        return result;
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueListTableInterface#executeFindQueueNames(java.sql.Connection)
     */
    public List<String> executeFindQueueNames(Connection connection) throws SQLException, IllegalArgumentException {
        ArrayList<String> names = new ArrayList<String>();
        logger.debug("executeFindQueueNames begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        
        try {
            pstmt = getPreparedStatement(findQueueNamesQuery, connection);
            rset = pstmt.executeQuery();

            if (rset != null) {
                while (rset.next()) {
                    String name = rset.getString(1);
                    names.add(name);
                }
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeFindQueueNames end");
        return names;
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueListTableInterface#executeTableCreation(java.lang.String,
     *      java.sql.Connection)
     */
    public void executeTableCreation(String queueName, Connection connection) throws SQLException, IllegalArgumentException {
        if (queueName == null) {
            throw new IllegalArgumentException("The queue name is null");
        }

        logger.debug("executeTableCreation begin");
        PreparedStatement pstmt = null;

        try {
            pstmt = getPreparedStatement(getTableCreationQuery(queueName), connection);
            pstmt.executeUpdate();

            logger.debug("table created for queue name=" + queueName);
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeTableCreation end");

    }

    public void executeTableDrop(String queueName, Connection connection) throws SQLException, IllegalArgumentException {
        if (queueName == null) {
            throw new IllegalArgumentException("The queue name is null");
        }

        logger.debug("executeTableDrop begin");
        PreparedStatement pstmt = null;

        try {
            pstmt = getPreparedStatement(getTableDropQuery(queueName), connection);
            pstmt.executeUpdate();

            logger.debug("table dropped for queue name=" + queueName);
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeTableDrop end");

    }

    /**
     * Returns the query for inserting.
     * 
     * @return The query for inserting.
     */
    protected Query makeInsertQuery() {
        StringBuffer query = new StringBuffer();
        query.append("insert into ");
        query.append(CommandQueueListTable.NAME_TABLE);
        query.append(" ( ");
        query.append(CommandQueueListTable.NAME);
        query.append(") values(?)");

        Query insertQuery = new Query();
        insertQuery.setName(Query.INSERT_STATEMENT);
        insertQuery.setStatement(query.toString());

        return insertQuery;
    }

    /**
     * Returns the query for deleting.
     * 
     * @return The query for deleting.
     */
    protected Query makeDeleteQuery() {
        StringBuffer query = new StringBuffer();
        query.append("delete from ");
        query.append(CommandQueueListTable.NAME_TABLE);
        query.append(" where ");
        query.append(CommandQueueListTable.NAME);
        query.append(" = ?");

        Query deleteQuery = new Query();
        deleteQuery.setName(Query.DELETE_STATEMENT);
        deleteQuery.setStatement(query.toString());

        return deleteQuery;
    }

    /**
     * Returns the query for checking if a queue name is registered.
     * 
     * @return The query for for checking if a queue name is registered.
     */
    protected Query makeIsQueueNamePresentQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select count(*) from ");
        query.append(CommandQueueListTable.NAME_TABLE);
        query.append(" where ");
        query.append(CommandQueueListTable.NAME);
        query.append(" = ?");

        Query isQueueNamePresentQuery = new Query();
        isQueueNamePresentQuery.setName(IS_QUEUE_NAME_PRESENT);
        isQueueNamePresentQuery.setStatement(query.toString());

        return isQueueNamePresentQuery;
    }

    /**
     * Returns the query for finding current queue names registered.
     * 
     * @return The query for finding current queue names registered.
     */
    protected Query makeFindQueueNamesQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select ");
        query.append(CommandQueueListTable.NAME);
        query.append(" from ");
        query.append(CommandQueueListTable.NAME_TABLE);

        Query findQueueNamesQueryQuery = new Query();
        findQueueNamesQueryQuery.setName(FIND_QUEUE_NAMES_QUERY);
        findQueueNamesQueryQuery.setStatement(query.toString());
        logger.debug("findScheduledCommandsQuery = " + query.toString());

        return findQueueNamesQueryQuery;
    }

    /**
     * Returns the query for a table creation.
     * 
     * @param queueName
     *            The queue name.
     * @return The query for a table creation.
     */
    protected Query getTableCreationQuery(String queueName) {
        StringBuffer query = new StringBuffer();
        query.append("CREATE TABLE IF NOT EXISTS ");
        query.append(queueName).append(" (");
        query.append(CommandQueueTable.ID_FIELD).append(" BIGINT NOT NULL AUTO_INCREMENT, ");
        query.append(CommandQueueTable.COMMAND_ID_FIELD).append(" BIGINT NOT NULL, ");
        query.append(CommandQueueTable.IS_SCHEDULED).append( " BOOL NOT NULL, ");
        query.append(CommandQueueTable.PRIORITY_LEVEL).append(" TINYINT UNSIGNED NOT NULL DEFAULT 0, ");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD).append(" VARCHAR(14) NULL, ");
        query.append(CommandQueueTable.EXECUTION_MODE_FIELD).append(" CHAR(1) NOT NULL, ");
        query.append(" PRIMARY KEY (").append(CommandQueueTable.ID_FIELD ).append("))");

        Query tableCreation = new Query();
        tableCreation.setName(TABLE_CREATION_QUERY);
        tableCreation.setStatement(query.toString());

        return tableCreation;
    }

    /**
     * Returns the query for a table drop.
     * 
     * @param queueName
     *            The queue name.
     * @return The query for a table drop.
     */
    protected Query getTableDropQuery(String queueName) {
        Query tableDrop = new Query();
        tableDrop.setName(TABLE_DROP_QUERY);
        tableDrop.setStatement("DROP  TABLE IF EXISTS " + queueName);

        return tableDrop;
    }
}
