/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *          Luigi Zangrando <luigi.zangrando@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.queue.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.AbstractTable;
import org.glite.ce.common.db.Query;
import org.glite.ce.creamapi.cmdmanagement.Command;
import org.glite.ce.creamapi.cmdmanagement.Command.ExecutionModeValues;

/**
 * Class representing the Data Access Object to a command queue table in the
 * database.
 * 
 */
public class CommandQueueTable extends AbstractTable implements CommandQueueTableInterface {
    /** The logger */
    private static final Logger logger = Logger.getLogger(CommandQueueTable.class);

    /** The table name: it changes according to the command queue name */
    private String nameTable = "job_management";

    /** Sequence name */
    private String sequenceName = "job_management_id_seq";

    // The table columns
    public final static String ID_FIELD = "id";
    public final static String COMMAND_ID_FIELD = "commandId";
    public final static String IS_SCHEDULED = "isScheduled";
    public final static String PRIORITY_LEVEL = "priorityLevel";
    public final static String COMMAND_GROUP_ID_FIELD = "commandGroupId";
    public final static String EXECUTION_MODE_FIELD = "executionMode";

    /** find first command in queue query name */
    public static final String FIND_FIRST_COMMAND_IN_QUEUE = "findFirstCommandInQueue";

    /** find scheduled jobs in queue query name */
    public static final String FIND_SCHEDULED_COMMANDS = "findScheduledCommands";

    /** select for update query name */
    public static final String SELECT_FOR_UPDATE = "selectForUpdateQuery";

    /** The find first item in queue query */
    protected Query findCommandIdInQueueQuery = null;

    /** The find scheduled commands in queue query */
    protected Query findScheduledCommandsQuery = null;

    protected Query getQueueSizeQuery = null;

    /**
     * Constructor.
     * 
     * @param queueName
     *            The name of the queue, which corresponds to the same one of
     *            the associated table.
     */
    public CommandQueueTable(String queueName) {
        if (queueName != null) {
            nameTable = queueName;
            sequenceName = queueName + "_id_seq";

            insertQuery = makeInsertQuery();
            updateQuery = makeUpdateQuery();
            deleteQuery = makeDeleteQuery();            
            getQueueSizeQuery = makeGetQueueSizeQuery();
            findScheduledCommandsQuery = makeFindScheduledCommandsQuery();
            findCommandIdInQueueQuery = makeFindCommandIdInQueueQuery();
        } else {
            throw new IllegalArgumentException("The name of queue is null");
        }
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeDelete(java.lang.String,
     *      java.sql.Connection)
     */
    public void executeDelete(String commandId, Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeDelete begin");

        if (commandId == null) {
            throw new IllegalArgumentException("The command id is null");
        }
        if (Long.parseLong(commandId) < 0) {
            throw new IllegalArgumentException("The id = " + commandId + " is invalid");
        }

        PreparedStatement pstmt = null;
        try {
            // Get the id from the sequence named sequenceName
            pstmt = getPreparedStatement(deleteQuery, connection);
            pstmt.setLong(1, Long.parseLong(commandId));
            
            logger.debug("delete statetement = " + pstmt.toString());

            int rowCount = pstmt.executeUpdate();
            logger.debug("deleted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeDelete end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeFindFirstCommandInQueue(java.sql.Connection)
     */
    public CommandQueueItem executeFindFirstCommandInQueue(int priorityLevel, Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeFindFirstCommandInQueue begin");
        List<CommandQueueItem> result = executeFindCommandInQueue(priorityLevel, 1, connection);
        logger.debug("executeFindFirstCommandInQueue end");
        
        if(result != null && result.size() > 0) {
            return result.get(0);
        }
        
        return null;
    }

    
    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeFindFirstCommandInQueue(java.sql.Connection)
     */
    public List<CommandQueueItem> executeFindCommandInQueue(int priorityLevel, int limit, Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeFindCommandInQueue begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        List<CommandQueueItem> result = new ArrayList<CommandQueueItem>(0);
        
        if (limit <= 0) {
            limit = 1;
        }
        
        try {
            // Get the first command in queue
            pstmt = getPreparedStatement(findCommandIdInQueueQuery, connection);
            pstmt.setInt(1, priorityLevel);
            pstmt.setInt(2, limit);

            rset = pstmt.executeQuery();

            if (rset != null) {
                CommandQueueItem item = null;  

                while (rset.next()) {
                    item = new CommandQueueItem();
                    item.setId(rset.getLong(1));
                    item.setCommandId(rset.getLong(2));
                    item.setCommandGroupId(rset.getString(3));
                    item.setScheduled(rset.getBoolean(4));
                    item.setPriorityLevel(rset.getInt(5));
                    item.setExecutionMode(ExecutionModeValues.PARALLEL.equals(rset.getString(6))? ExecutionModeValues.PARALLEL : ExecutionModeValues.SERIAL);

                    result.add(item);
                }
                
                logger.debug("found " + result.size() + " commands!");
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        
        logger.debug("executeFindCommandInQueue end");        
        return result;
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeFindScheduledCommands(java.sql.Connection)
     */
    public List<CommandQueueItem> executeFindScheduledCommands(Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeFindScheduledJobs begin");

        ArrayList<CommandQueueItem> commands = new ArrayList<CommandQueueItem>(0);
        PreparedStatement pstmt = null;
        ResultSet rset = null;

        try {
            // Get the scheduled command list
            pstmt = getPreparedStatement(findScheduledCommandsQuery, connection);

            rset = pstmt.executeQuery();

            CommandQueueItem item = null;

            if (rset != null) {
                while (rset.next()) {
                    item = new CommandQueueItem();
                    item.setId(rset.getLong(1));
                    item.setCommandId(rset.getLong(2));
                    item.setCommandGroupId(rset.getString(3));
                    item.setScheduled(rset.getBoolean(4));
                    item.setPriorityLevel(rset.getInt(5));
                    item.setExecutionMode(ExecutionModeValues.PARALLEL.equals(rset.getString(6))? ExecutionModeValues.PARALLEL : ExecutionModeValues.SERIAL);

                    commands.add(item);
                }
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        
        logger.debug("executeFindScheduledJobs end");
        return commands;
    }

    public int executeGetQueueSize(Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeFindFirstCommandInQueue begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        int queuesize = 0;

        try {
            // Get the scheduled command list
            pstmt = getPreparedStatement(getQueueSizeQuery, connection);

            rset = pstmt.executeQuery();

            if (rset != null) {
                rset.next();

                if (rset.getRow() > 0) {
                    queuesize = rset.getInt(1);
                } else {
                    logger.debug("No command in queue");
                }
            }
            logger.debug("queue size is " + queuesize);
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        
        logger.debug("executeFindFirstCommandInQueue end");
        return queuesize;
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeInsert(org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueItem,
     *      java.sql.Connection)
     */
    public void executeInsert(CommandQueueItem item, Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeInsert begin");

        if (item == null) {
            throw new IllegalArgumentException("The command queue item is null");
        }
        // Check the non nullable fields
        if (item.getCommandId() < 0) {
            throw new IllegalArgumentException("The command queue item has an invalid command id=" + item.getCommandId());
        }
        if (item.getPriorityLevel() < Command.LOW_PRIORITY || item.getPriorityLevel() > Command.HIGH_PRIORITY) {
            throw new IllegalArgumentException("The command queue item has an invalid priority level=" + item.getPriorityLevel());
        }
        if (item.getExecutionMode() == Command.ExecutionModeValues.SERIAL && (item.getCommandGroupId() == null || ("".equals(item.getCommandGroupId())))) {
            logger.error("CommandGroupId field mustn't be empty for serial commands.");
            throw new IllegalArgumentException("CommandGroupId field mustn't be empty for serial commands.");
        }

        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            // Get the id from the sequence named sequenceName
            pstmt = getPreparedStatement(insertQuery, connection);
            pstmt.setLong(1, item.getCommandId());
            pstmt.setBoolean(2, false);
            pstmt.setInt(3, item.getPriorityLevel());
            pstmt.setString(4, item.getCommandGroupId());
            pstmt.setString(5, item.getExecutionMode().getStringValue());
            logger.debug("insert statetement = " + pstmt.toString());

            int rowCount = pstmt.executeUpdate();

            // Get autogenerated keys
            long id = -1;
            rset = null;
            rset = pstmt.getGeneratedKeys();

            if (rset.next()) {
                id = rset.getLong(1);
                item.setId(id);
            } else {
                throw new SQLException("Problem in retrieving autogenerated keys");
            }
            logger.debug("inserted " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeInsert end");
    }

    /**
     * @see org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueTableInterface#executeUpdate(org.glite.ce.cream.cmdmanagement.queue.db.CommandQueueItem,
     *      java.sql.Connection)
     */
    public void executeUpdate(CommandQueueItem item, Connection connection) throws SQLException, IllegalArgumentException {
        logger.debug("executeUpdate begin");

        if (item == null) {
            throw new IllegalArgumentException("The command queue item is null");
        }
        if (item.getCommandId() < 0) {
            throw new IllegalArgumentException("The command id = " + item.getCommandId() + " is invalid");
        }

        PreparedStatement pstmt = null;
        try {
            // Get the id from the sequence named sequenceName
            pstmt = getPreparedStatement(updateQuery, connection);
            pstmt.setBoolean(1, item.isScheduled());
            pstmt.setLong(2, item.getCommandId());
            logger.debug("update statetement = " + pstmt.toString());

            int rowCount = pstmt.executeUpdate();
            logger.debug("updated " + rowCount + " rows");
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeUpdate end");
    }

    /**
     * Returns the query for deleting.
     * 
     * @return The query for deleting.
     */
    protected Query makeDeleteQuery() {
        StringBuffer query = new StringBuffer();
        query.append("delete from ");
        query.append(this.nameTable);
        query.append(" where ");
        query.append(COMMAND_ID_FIELD);
        query.append(" = ");
        query.append(" ? ");

        Query deleteQuery = new Query();
        deleteQuery.setName(Query.DELETE_STATEMENT);
        deleteQuery.setStatement(query.toString());

        return deleteQuery;
    }

    /**
     * Returns the query for inserting.
     * 
     * @return The query for inserting.
     */
    protected Query makeFindCommandIdInQueueQuery() {
        String nameTable2 = nameTable + "2";      
        
        StringBuffer query = new StringBuffer();            
        query.append("select ");
        query.append(CommandQueueTable.ID_FIELD).append(", ");
        query.append(CommandQueueTable.COMMAND_ID_FIELD).append(", ");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD).append(", ");
        query.append(CommandQueueTable.IS_SCHEDULED).append(", ");
        query.append(CommandQueueTable.PRIORITY_LEVEL).append(", ");
        query.append(CommandQueueTable.EXECUTION_MODE_FIELD);
        query.append(" from ");
        query.append(nameTable);
        query.append(" where ");
        query.append(CommandQueueTable.PRIORITY_LEVEL);
        query.append(" = ? and ");
        query.append(CommandQueueTable.IS_SCHEDULED);
        query.append(" = false and ");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD);
        query.append(" NOT IN (select ");
        query.append(nameTable2);
        query.append(".");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD);
        query.append(" from ");
        query.append(nameTable);
        query.append(" ");
        query.append(nameTable2);
        query.append(" where ");
        query.append(nameTable2);
        query.append(".");
        query.append(CommandQueueTable.IS_SCHEDULED);
        query.append(" = true and ");
        query.append(nameTable2);
        query.append(".");
        query.append(CommandQueueTable.EXECUTION_MODE_FIELD);
        query.append(" = '");
        query.append(Command.EXECUTION_MODE_SERIAL);
        query.append("') order by ");
        query.append(CommandQueueTable.ID_FIELD);
        query.append(" ASC limit ?"); // for update");

        Query findFirstItemInQueueQuery = new Query();
        findFirstItemInQueueQuery.setName(FIND_FIRST_COMMAND_IN_QUEUE);
        findFirstItemInQueueQuery.setStatement(query.toString());
        logger.debug("findFirstItemInQueueQuery = " + query.toString());

        return findFirstItemInQueueQuery;
    }

    /**
     * Returns the query for finding currently scheduled commands.
     * 
     * @return The query for finding currently scheduled commands.
     */
    protected Query makeFindScheduledCommandsQuery() {
        StringBuffer query = new StringBuffer();
        query.append("select ");
        query.append(CommandQueueTable.ID_FIELD).append(", ");
        query.append(CommandQueueTable.COMMAND_ID_FIELD).append(", ");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD).append(", ");
        query.append(CommandQueueTable.IS_SCHEDULED).append(", ");
        query.append(CommandQueueTable.PRIORITY_LEVEL).append(", ");
        query.append(CommandQueueTable.EXECUTION_MODE_FIELD);
        query.append(" from ");
        query.append(this.nameTable);
        query.append(" where ");
        query.append(CommandQueueTable.IS_SCHEDULED);
        query.append(" = true");

        Query findScheduledCommandsQuery = new Query();
        findScheduledCommandsQuery.setName(FIND_SCHEDULED_COMMANDS);
        findScheduledCommandsQuery.setStatement(query.toString());
        logger.debug("findScheduledCommandsQuery = " + query.toString());

        return findScheduledCommandsQuery;
    }

    /**
     * Returns the query for inserting.
     * 
     * @return The query for inserting.
     */
    protected Query makeInsertQuery() {
        StringBuffer query = new StringBuffer();
        query.append("insert into ");
        query.append(this.nameTable);
        query.append(" ( ");
        query.append(CommandQueueTable.COMMAND_ID_FIELD + ", ");
        query.append(CommandQueueTable.IS_SCHEDULED + ", ");
        query.append(CommandQueueTable.PRIORITY_LEVEL + ", ");
        query.append(CommandQueueTable.COMMAND_GROUP_ID_FIELD + ", ");
        query.append(CommandQueueTable.EXECUTION_MODE_FIELD);
        query.append(") values(?, ?, ?, ?, ?)");
        
        Query insertQuery = new Query();
        insertQuery.setName(Query.INSERT_STATEMENT);
        insertQuery.setStatement(query.toString());

        return insertQuery;
    }

    protected Query makeGetQueueSizeQuery() {
        Query getQueueSizeQuery = new Query();
        getQueueSizeQuery.setName(Query.SELECT_STATEMENT);
        getQueueSizeQuery.setStatement("select count(1) from " + nameTable + " where isScheduled=false");

        return getQueueSizeQuery;
    }

    /**
     * @return the nameTable
     */
    public String getNameTable() {
        return nameTable;
    }

    /**
     * @return the sequenceName
     */
    public String getSequenceName() {
        return sequenceName;
    }

    /**
     * Returns the query string for updating.
     * 
     * @param cmd
     *            The Command to be updated.
     * @return The query string for updating.
     */
    protected Query makeUpdateQuery() {
        StringBuffer query = new StringBuffer();
        query.append("update ");
        query.append(this.nameTable);
        query.append(" set ");
        query.append(IS_SCHEDULED);
        query.append(" = ? where ");
        query.append(COMMAND_ID_FIELD);
        query.append(" = ?");
        
        Query updateQuery = new Query();
        updateQuery.setName(Query.UPDATE_STATEMENT);
        updateQuery.setStatement(query.toString());

        return updateQuery;
    }

    /**
     * @param nameTable
     *            the nameTable to set
     */
    public void setNameTable(String nameTable) {
        this.nameTable = nameTable;
    }

    /**
     * @param sequenceName
     *            the sequenceName to set
     */
    public void setSequenceName(String sequenceName) {
        this.sequenceName = sequenceName;
    }
}
