/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Silvano Squizzato <silvano.squizzato@pd.infn.it>
 *
 */

package org.glite.ce.cream.cmdmanagement.queue.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;

import org.glite.ce.common.db.TableInterface;

/**
 * Interface to access the command table in the database.
 *
 */
public interface CommandQueueTableInterface extends TableInterface {
    
    /**
     * Inserts a entry referring to a command identified by a given id into the command queue.
     * @param item The command item to be inserted.
     * @connection The connection.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public void executeInsert(CommandQueueItem item, Connection connection) throws SQLException, IllegalArgumentException;
    
    /**
     * Deletes the entry referring to a command identified by <code>commandId</code> from the command queue.
     * @param commandId The id of the command to be deleted.
     * @connection The connection.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public void executeDelete(String commandId, Connection connection) throws SQLException, IllegalArgumentException;
    
    /**
     * Updates an item in the command queue.
     * The update is based on command id.
     * @param item The command item to be updated.
     * @connection The connection.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public void executeUpdate(CommandQueueItem item, Connection connection) throws SQLException, IllegalArgumentException;
        
    /**
     * Returns, if it exists, the first unscheduled command having the specified priorityLevel from the command queue.
     * Returns null if the queue has got no unscheduled command.
     * @param priorityLevel The priorityLevel.
     * @connection The connection.
     * @return Returns, if it exists, the first unscheduled command having the specified priorityLevel from the command queue.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public CommandQueueItem executeFindFirstCommandInQueue(int priorityLevel, Connection connection) throws SQLException, IllegalArgumentException;
  
    /**
     * Returns, if it exists, the first N unscheduled commands having the specified priorityLevel from the command queue.
     * Returns null if the queue has got no unscheduled command.
     * @param priorityLevel The priorityLevel.
     * @param limit The max number of results.
     * @connection The connection.
     * @return Returns, if it exists, the first N unscheduled commands having the specified priorityLevel from the command queue.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public List<CommandQueueItem> executeFindCommandInQueue(int priorityLevel, int limit, Connection connection) throws SQLException, IllegalArgumentException;
    
 
    /**
     * Returns the list of commands scheduled at a given time.
     * If no matching occurs the returned list is empty. 
     * @connection The connection.
     * @return Returns the list of commands scheduled at a given time.
     * If no matching occurs the returned list is empty.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public List<CommandQueueItem> executeFindScheduledCommands(Connection connection) throws SQLException, IllegalArgumentException;
    

    /**
     * Returns the queue size.
     * @connection The connection.
     * @return Returns the queue size.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public int executeGetQueueSize(Connection connection) throws SQLException, IllegalArgumentException;
}
