/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
package org.glite.ce.cream.configuration;

import java.util.ArrayList;
import java.util.List;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.NamingEnumeration;
import javax.naming.directory.DirContext;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchResult;

import org.glite.ce.cream.configuration.CREAMConfigConstants.ServiceAttribute;

import org.glite.ce.commonj.utils.CEUtils;

import org.apache.log4j.Logger;

public class ServiceConfig {

    private static Logger logger = Logger.getLogger(ServiceConfig.class.getName());

    private static ServiceConfig serviceConfiguration = null;

    protected List<Property> properties;
    protected List<CommandExecutorConfig> commandExecutor;

    protected DirContext rootContext;

    public ServiceConfig(String confProvider, String confProviderURL)
        throws NamingException {

        if(confProvider == null) {
            throw new NamingException("Configuration provider not defined!");
        }

        if(confProviderURL == null) {
            throw new NamingException("Configuration provider URL not defined!");
        }

        Hashtable<String, String> jndiProperties = new Hashtable<String, String>(2);
        jndiProperties.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
        jndiProperties.put(Context.PROVIDER_URL, confProviderURL);

        rootContext = (DirContext) new InitialDirContext(jndiProperties);

    }

    public String getStringParameter(String name){
        try{
            Attribute attribute = rootContext.getAttributes("").get(name);
            if( attribute!=null )
                return (String) attribute.get(0);
            else
                logger.debug("Cannot find attribute " + name);
                
        }catch(Exception ex){
            logger.debug(ex.getMessage());
        }

        return null;
    }
    
    public String getStringParameter(ServiceAttribute name){
    	return this.getStringParameter(name.toString());
    }

    public int getIntParameter(String name, int defValue){
        try{
            Attribute attribute = rootContext.getAttributes("").get(name);
            if( attribute!=null )
                return Integer.parseInt((String)attribute.get(0));
            else
                logger.debug("Cannot find attribute " + name);
        } catch (Exception ex) {
            logger.error(ex.getMessage());
        }

        return defValue;
    }
    
    public int getIntParameter(ServiceAttribute name, int defValue){
    	return this.getIntParameter(name.toString(), defValue);
    }

    public List<CommandExecutorConfig> getCommandExecutorList(){

        ArrayList<CommandExecutorConfig> result =
            new ArrayList<CommandExecutorConfig>();

        try{
            BasicAttributes queryAttrs = new BasicAttributes("category", "commandexecutor");

            NamingEnumeration<SearchResult> allExec = rootContext.search("", queryAttrs);
            while( allExec.hasMoreElements() ) {
                SearchResult tmpExec = allExec.nextElement();
                result.add((CommandExecutorConfig)tmpExec.getObject());
            }
        }catch(NamingException ex){
            logger.debug(ex.getMessage());
        }

        return result;
    }




/*
 * TODO remove this section of code
 */
    public ServiceConfig() {
        properties = new ArrayList(0);
        commandExecutor = new ArrayList(0);
    }

    public List<Property> getProperties() {
        return properties;
    }

    public void setProperties(List<Property> properties) {
        this.properties.clear();
        this.properties.addAll(properties);
    }

    public void addProperty(Property property) {
        if (property == null) {
            return;
        }
        properties.add(property);
    }

    public List<CommandExecutorConfig> getCommandExecutor() {
        return commandExecutor;
    }

    public void setCommandExecutor(List<CommandExecutorConfig> commandExecutor) {
        this.commandExecutor.clear();
        this.commandExecutor.addAll(commandExecutor);
    }

    public void addCommandExecutor(CommandExecutorConfig commandExecutorProperty) {
        if (commandExecutorProperty == null) {
            return;
        }
        commandExecutor.add(commandExecutorProperty);
    }
/*
 * end section
 */



    public static ServiceConfig getConfiguration(){
        if( serviceConfiguration==null ){

            String confProvider = CEUtils.getConfigurationProvider(CEUtils.CREAM_LABEL);;
            String confProviderURL = CEUtils.getConfigurationURL(CEUtils.CREAM_LABEL);

            try{
            	logger.debug("Creating service configuration: " + confProviderURL);
                serviceConfiguration = new ServiceConfig(confProvider, confProviderURL);
            }catch(NamingException nEx){
                logger.error(nEx.getMessage(), nEx);
            }
        }
        return serviceConfiguration;
    }

    public static void main(String[] args){
        try{

            String confProvider = "org.glite.ce.cream.configuration.basic.CREAMConfigContextFactory";
            String confProviderURL = args[0];

            serviceConfiguration = new ServiceConfig(confProvider, confProviderURL);

            System.out.println(ServiceAttribute.JOB_DIR_LABEL + ": " +
                               serviceConfiguration.getStringParameter(ServiceAttribute.JOB_DIR_LABEL));
            System.out.println(ServiceAttribute.DELEGATION_KEYSIZE_LABEL + ": " +
            	               serviceConfiguration.getIntParameter(ServiceAttribute.DELEGATION_KEYSIZE_LABEL, 1024));
            System.out.println(ServiceAttribute.DELEGATION_STORAGE_DB_LABEL + ": " +
 	               serviceConfiguration.getStringParameter(ServiceAttribute.DELEGATION_STORAGE_DB_LABEL));

            List<CommandExecutorConfig> execList = serviceConfiguration.getCommandExecutorList();
            for(int k=0; k<execList.size(); k++){
                CommandExecutorConfig executor = execList.get(k);
                System.out.println(executor.getName() + " " +
                                   executor.getCategory() + " " +
                                   executor.getJarFileName());
                List<Property> propList = executor.getProperties();
                for(int j=0; j<propList.size(); j++){
                	Property prop = propList.get(j);
                	System.out.println("    " + prop.getKey() + " " + prop.getValue());
                }
            }
            

        }catch(Throwable th){
            th.printStackTrace();
        }
    }

}
