/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.cream.configuration.basic;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigNodes;
import org.glite.ce.commonj.configuration.basic.ConfigParser;
import org.glite.ce.commonj.configuration.basic.TriggerManager;
import org.glite.ce.cream.configuration.CommandExecutorConfig;
import org.glite.ce.cream.configuration.Property;
import org.glite.ce.cream.configuration.CREAMConfigConstants.ServiceAttribute;
import org.xml.sax.Attributes;
import org.xml.sax.SAXParseException;

public class CREAMConfigParser implements ConfigParser {
    private static final String SERVICE_LABEL = "service";
    private static final String CMD_EXECUTOR_LABEL = "commandexecutor";

    protected TriggerManager triggerManager;
    protected ArrayList<Property> paramList;

    protected String[] serviceAttributes;

    public CREAMConfigParser(TriggerManager tMan) {
        triggerManager = tMan;

        ServiceAttribute[] attrList = ServiceAttribute.values();
        serviceAttributes = new String[attrList.length];
        
        for (int k = 0; k < attrList.length; k++) {
            serviceAttributes[k] = attrList[k].toString();
        }
    }

    public String getObjectName(String uri, String name, String qName, Attributes attributes) throws SAXParseException {
        if (qName.equals(SERVICE_LABEL) || qName.equals(CMD_EXECUTOR_LABEL)) {
            String result = attributes.getValue("id");
            if (result == null)
                throw new SAXParseException("Missing service name", null);

            return result;
        }

        return null;
    }

    public void startElement(String uri, String name, String qName, Attributes attributes, ConfigItem prevObj) throws SAXParseException {
        if (qName.equals(SERVICE_LABEL)) {

            Long tmpl = new Long(System.currentTimeMillis());
            boolean configChanged = false;

            ConfigNodes currentList = (ConfigNodes) prevObj.getContent();
            
            if (currentList == null) {
                currentList = new ConfigNodes();
                prevObj.setContent(currentList);
                prevObj.put("creationtime", tmpl);
                prevObj.put("category", SERVICE_LABEL);
            }

            for (int k = 0; k < serviceAttributes.length; k++) {
                String tmp1 = attributes.getValue(serviceAttributes[k]);
                String tmp2 = (String) prevObj.get(serviceAttributes[k]);

                if (tmp1 == null) {
                    if (tmp2 != null) {
                        prevObj.remove(serviceAttributes[k]);
                        configChanged = true;
                    }
                } else if (tmp2 == null || !tmp1.equalsIgnoreCase(tmp2)) {
                    prevObj.put(serviceAttributes[k], tmp1);
                    configChanged = true;
                }
            }

            if (configChanged) {
                prevObj.put("modificationtime", tmpl);
            }
        } else if (qName.equals(CMD_EXECUTOR_LABEL)) {
            Long tmpl = new Long(System.currentTimeMillis());
            paramList = new ArrayList<Property>();

            CommandExecutorConfig currentExecConfig = null;
            if (prevObj.getContent() == null) {
                currentExecConfig = new CommandExecutorConfig();
                prevObj.setContent(currentExecConfig);
                prevObj.put("creationtime", tmpl);
                prevObj.put("category", CMD_EXECUTOR_LABEL);
            } else {
                currentExecConfig = (CommandExecutorConfig) prevObj.getContent();
            }

            String tmps = attributes.getValue("filename");
            if (tmps == null) {
                throw new SAXParseException("Missing resource filename", null);
            }
            
            File newJarFile = new File(tmps);

            String oldFilename = currentExecConfig.getJarFileName();
            
            if (!tmps.equals(oldFilename)) {
                currentExecConfig.setJarFileName(tmps);
                prevObj.put("modificationtime", tmpl);

                if (oldFilename != null)
                    triggerManager.removeTrigger(new File(oldFilename));
                triggerManager.addTrigger(newJarFile);
            } else {
                long jarLastModTime = newJarFile.lastModified();
                Object tmpo = prevObj.get("modificationtime");
                
                if (tmpo == null || ((Long) tmpo).longValue() < jarLastModTime) {
                    prevObj.put("modificationtime", tmpl);       
                }
            }

            tmps = attributes.getValue("category");
            
            if (tmps == null) {
                throw new SAXParseException("Missing resource category", null);
            }
            
            String oldCategory = currentExecConfig.getCategory();
            
            if (!tmps.equals(oldCategory)) {
                currentExecConfig.setCategory(tmps);
                prevObj.put("modificationtime", tmpl);
            }

            tmps = attributes.getValue("id");
            
            if (tmps == null) {
                throw new SAXParseException("Missing resource id", null);
            }
            
            String oldName = currentExecConfig.getName();
            if (!tmps.equals(oldName)) {
                currentExecConfig.setName(tmps);
                prevObj.put("modificationtime", tmpl);
            }

        } else if (qName.equals("parameter")) {
            String pName = attributes.getValue("name");
            String pValue = attributes.getValue("value");
            
            if (pName == null || pValue == null) {
                throw new SAXParseException("Bad parameter definition", null);
            }
                
            paramList.add(new Property(pName.trim(), pValue.trim()));
        }
    }

    public void endElement(String uri, String name, String qName, ConfigItem currentObj) {
        if (qName.equals(CMD_EXECUTOR_LABEL)) {
            boolean configChanged = false;

            CommandExecutorConfig currentExecConfig = (CommandExecutorConfig) currentObj.getContent();

            List oldParamList = currentExecConfig.getProperties();

            if (oldParamList.size() != paramList.size()) {
                configChanged = true;
            } else {
                int nof = 0;
                for (int k = 0; k < paramList.size(); k++) {
                    for (int j = 0; j < oldParamList.size(); j++) {
                        if (paramList.get(k).equals(oldParamList.get(j))) {
                            nof++;
                        }
                    }
                }

                configChanged = (nof != oldParamList.size());
            }

            if (configChanged) {
                currentExecConfig.setProperties(paramList);
                currentObj.put("modificationtime", new Long(System.currentTimeMillis()));
            }
        }
    }

    public void characters(char[] chars, int start, int length, ConfigItem currentContext) {
    }
}

