/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.cream.delegationmanagement;

import java.util.Calendar;
import java.util.List;
import java.util.Timer;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.cream.configuration.ServiceConfig;
import org.glite.ce.cream.configuration.CREAMConfigConstants.ServiceAttribute;
import org.glite.ce.cream.dbmanagement.DBInfoManager;
import org.glite.ce.cream.delegationmanagement.db.DelegationDBManager;
import org.glite.ce.creamapi.delegationmanagement.Delegation;
import org.glite.ce.creamapi.delegationmanagement.DelegationException;
import org.glite.ce.creamapi.delegationmanagement.DelegationManagementException;
import org.glite.ce.creamapi.delegationmanagement.DelegationManagerInterface;

public class DelegationManager implements DelegationManagerInterface {
    private final static Logger logger = Logger.getLogger(DelegationManager.class.getName());
    private static Timer timer = null;
    private static DelegationManager delegationManager = null;
    private static DelegationDBManager delegationDBManager = null;

    
    public static DelegationManager getInstance() throws DelegationManagementException {
        if (delegationManager == null) {
            delegationManager = new DelegationManager();
            delegationManager.init();
        }

        return delegationManager;
    }

    public void deleteDelegation(Delegation delegation) throws DelegationException, DelegationManagementException {
        if (delegation == null) {
            throw new IllegalArgumentException("delegation not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }

        try {
            deleteDelegation(delegation.getId(), delegation.getDN(), delegation.getLocalUser());
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public void deleteDelegation(String delegationId, String dn, String localUser) throws DelegationException, DelegationManagementException {
        if (delegationId == null || delegationId.length() == 0) {
            throw new IllegalArgumentException("delegationId not specified!");
        }

        if (dn == null || dn.length() == 0) {
            throw new IllegalArgumentException("user DN not specified!");
        }

        if (localUser == null || localUser.length() == 0) {
            throw new IllegalArgumentException("localUser not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }

        try {
            delegationDBManager.deleteDelegation(delegationId, dn, localUser);
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public void destroy() {
        if (timer != null) {
            timer.cancel();
            timer = null;
        }
    }

    public Delegation getDelegation(String delegationId, String dn, String localUser) throws DelegationException, DelegationManagementException {
        if (delegationId == null || delegationId.length() == 0) {
            throw new IllegalArgumentException("delegationId not specified!");
        }

        if (dn == null || dn.length() == 0) {
            throw new IllegalArgumentException("user DN not specified!");
        }

        if (localUser == null || localUser.length() == 0) {
            throw new IllegalArgumentException("localUser not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }
        
        try {
            return delegationDBManager.getDelegation(delegationId, dn, localUser);
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public String getDelegationSuffix() throws DelegationException, DelegationManagementException {
        try {
            //return DBInfoManager.getDelegationSuffix(DelegationManagerInterface.DELEGATION_DATASOURCE_NAME);
            return DBInfoManager.getDelegationSuffix("java:comp/env/jdbc/creamdb");
        } catch(IllegalArgumentException ex) {
            throw new DelegationManagementException(ex.getMessage());
        }
    }    
    
    public List<Delegation> getExpiredDelegation() throws DelegationException, DelegationManagementException {
        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }
        
        try {
            return delegationDBManager.getExpiredDelegation();
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public void init() throws DelegationManagementException {
        logger.info("Initializing the DelegationManager...");
        try {
            delegationDBManager = new DelegationDBManager(DelegationManagerInterface.DELEGATION_DATASOURCE_NAME);

            ServiceConfig serviceConf = ServiceConfig.getConfiguration();

            int purgerRateInMinutes = 720;

            try {
                purgerRateInMinutes = Integer.parseInt(serviceConf.getStringParameter(ServiceAttribute.DELEGATION_PURGE_RATE_LABEL));
            } catch (Throwable t) {
                logger.warn("Configuration warning: wrong value for " + ServiceAttribute.DELEGATION_STORAGE_LABEL + " property => using default " + purgerRateInMinutes);
            }

            if (purgerRateInMinutes == -1) {
                logger.info("the delegation proxy purger is disabled!");
            } else {
                if (timer == null) {
                    Calendar time = Calendar.getInstance();
                    time.add(Calendar.SECOND, 10);

                    timer = new Timer("TIMER", true);
                    timer.schedule(new DelegationPurger(), time.getTime(), purgerRateInMinutes * 60000);
                }

                logger.info("Initializing the DelegationManager... DONE!");
            }
        } catch (DatabaseException e) {
            logger.error(e.getMessage());

            destroy();
            throw new DelegationManagementException(e.getMessage());
        }
    }    
    
    public void insertDelegation(Delegation delegation) throws DelegationException, DelegationManagementException {
        if (delegation == null) {
            throw new IllegalArgumentException("delegation not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }
        
        try {
            delegationDBManager.insertDelegation(delegation);
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public List<Delegation> listDelegations(String dn, String localUser) throws DelegationException, DelegationManagementException {
        if (dn == null || dn.length() == 0) {
            throw new IllegalArgumentException("user DN not specified!");
        }

        if (localUser == null || localUser.length() == 0) {
            throw new IllegalArgumentException("localUser not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }
        
        try {
            return delegationDBManager.listDelegations(dn, localUser);
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }

    public void updateDelegation(Delegation delegation) throws DelegationException, DelegationManagementException {
        if (delegation == null) {
            throw new DelegationException("delegation not specified!");
        }

        if (delegationDBManager == null) {
            throw new DelegationManagementException("DelegationDBManager not instantiated");
        }
        
        try {
            delegationDBManager.updateDelegation(delegation);
        } catch (DelegationException e) {
            logger.error(e.getMessage());
            throw e;
        }
    }
}

