package org.glite.ce.cream.delegationmanagement.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.common.db.DatasourceManager;
import org.glite.ce.commonj.utils.CEUtils;
import org.glite.ce.creamapi.delegationmanagement.Delegation;
import org.glite.ce.creamapi.delegationmanagement.DelegationException;
import org.glite.ce.creamapi.delegationmanagement.db.DelegationDBInterface;
import org.glite.security.util.DNHandler;

public class DelegationDBManager implements DelegationDBInterface {
    /** The logger */
    private final static Logger logger = Logger.getLogger(DelegationDBManager.class);
    private String datasourceName;

    public DelegationDBManager(String datasourceName) throws DatabaseException {
        if (datasourceName == null) {
            throw new IllegalArgumentException("datasourceName not specified!");
        }

        this.datasourceName = datasourceName;
    }

    public void deleteDelegation(String delegationId, String dn, String localUser) throws DelegationException {
        if (delegationId == null || delegationId.length() == 0) {
            throw new DelegationException("delegationId not specified!");
        }

        if (dn == null || dn.length() == 0) {
            throw new DelegationException("user DN not specified!");
        }

        if (localUser == null || localUser.length() == 0) {
            throw new DelegationException("localUser not specified!");
        }

        logger.debug("Begin deleteDelegation for delegationId=" + delegationId + " dn=" + dn + " localUser=" + localUser);

        Connection connection = getConnection();

        try {
            PreparedStatement deletePreparedStatement = null;
            String dnX500 = DNHandler.getDN(dn).getX500();
            try {
                deletePreparedStatement = connection.prepareStatement("delete from t_credential where dlg_id = ? and dn = ? and local_user = ?");
                deletePreparedStatement.setString(1, delegationId);
                deletePreparedStatement.setString(2, dnX500);
                deletePreparedStatement.setString(3, localUser);                
                deletePreparedStatement.executeUpdate();

                deletePreparedStatement = connection.prepareStatement("delete from t_credential_cache where dlg_id = ? and dn= ?");
                deletePreparedStatement.setString(1, delegationId);
                deletePreparedStatement.setString(2, dnX500);               
                deletePreparedStatement.executeUpdate();

                connection.commit();
            } catch (SQLException sqle) {
                logger.error("Failure on db interaction", sqle);
                try {
                    connection.rollback();
                } catch (SQLException e) {
                    logger.error(e);
                }
                throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
            } finally {
                if (deletePreparedStatement != null) {
                    try {
                        deletePreparedStatement.close();
                    } catch (SQLException sqle1) {
                        logger.error(sqle1);
                    }
                }

                if (connection != null) {
                    try {
                        connection.close();
                    } catch (SQLException e) {
                        logger.error(e);
                    }
                }
            }
        } catch (NullPointerException e) {
            throw new DelegationException("database connection problem");
        } catch (Throwable e) {
            throw new DelegationException("deleteDelegation error: " + e.getMessage());
        }

        logger.debug("End deleteDelegationProxy for delegationId=" + delegationId + " dn=" + dn + " localUser=" + localUser);
    }

    private Delegation makeDelegation(ResultSet rs) throws DelegationException, IllegalArgumentException  {
        if (rs == null) {
            throw new IllegalArgumentException("rs new defined!");            
        }
        
        try {
            StringTokenizer st = new StringTokenizer(rs.getString("voms_attrs"), "\t");
            List<String> vomsAttributes = new ArrayList<String>(0);

            while (st.hasMoreTokens()) {
                vomsAttributes.add(st.nextToken());
            }

            Delegation delegation = Delegation.getInstance(rs.getString("dlg_id"));
            delegation.setDN(rs.getString("dn"));
            delegation.setFQAN(rs.getString("fqan"));
            delegation.setVO(rs.getString("vo"));
            delegation.setLocalUser(rs.getString("local_user"));
            delegation.setLocalUserGroup(rs.getString("local_user_group"));
            delegation.setProxy(rs.getString("proxy"));
            delegation.setInfo(rs.getString("info"));
            delegation.setExpirationTime(rs.getTimestamp("termination_time"));
            delegation.setLastUpdateTime(rs.getTimestamp("last_update_time"));
            delegation.setStartTime(rs.getTimestamp("start_time"));
            delegation.setVOMSAttributes(vomsAttributes);

            return delegation;
        } catch (SQLException sqle) {
            logger.error("makeDelegation: failure on ResultSet: ", sqle);
            throw new DelegationException("Failure on ResultSet: " + sqle.getMessage());
        }
    }
    
    /**
     * Returns the connection to the database.
     * 
     * @return The connection to the database.
     * @throws DatabaseException
     */
    private Connection getConnection() throws DelegationException {
        Connection connection = null;
/*
        try {
            Class.forName("org.gjt.mm.mysql.Driver");
            connection = DriverManager.getConnection("jdbc:mysql://cream-10.pd.infn.it/delegationdb?user=cremino&password=creamtest");
            connection.setAutoCommit(false);
        } catch (ClassNotFoundException e) {
            throw new DelegationException("cannot get the database connection");
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
*/
        try {
            connection = DatasourceManager.getConnection(datasourceName);
        } catch (DatabaseException ex) {
            throw new DelegationException(ex.getMessage());
        }

        if (connection == null) {
            logger.error("cannot get the database connection");
            throw new DelegationException("cannot get the database connection");
        }

        return connection;
    }

    public Delegation getDelegation(String delegationId, String dn, String localUser) throws DelegationException {
        if (delegationId == null || delegationId.length() == 0) {
            throw new DelegationException("delegationId not specified!");
        }

        if (dn == null || dn.length() == 0) {
            throw new DelegationException("user DN not specified!");
        }

        if (localUser == null || localUser.length() == 0) {
            throw new DelegationException("localUser not specified!");
        }

        logger.debug("Begin getDelegation for delegationId=" + delegationId + " dn=" + dn + " localUser=" + localUser);

        Connection connection = getConnection();

        final String sql = "SELECT dlg_id, dn, fqan, vo, local_user, local_user_group, proxy, info, voms_attrs, termination_time, last_update_time, start_time FROM t_credential WHERE dlg_id = ? AND dn = ? AND local_user = ?";
        Delegation delegation = null;
        PreparedStatement getProxyStatement = null;
        ResultSet rs = null;

        try {
            getProxyStatement = connection.prepareStatement(sql);
            getProxyStatement.setString(1, delegationId);
            getProxyStatement.setString(2, DNHandler.getDN(dn).getX500());
            getProxyStatement.setString(3, localUser);
            
            rs = getProxyStatement.executeQuery();

            if (rs.next()) {
                delegation = makeDelegation(rs);                
            } else {
                throw new DelegationException("Delegation " + delegationId + " [dn=" + dn + "; localUser=" + localUser + "] not found!");        
            }
        } catch (SQLException sqle) {
            logger.error("Failure on db interaction", sqle);
            throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
        } finally {
            if (getProxyStatement != null) {
                try {
                    getProxyStatement.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }

            if (rs != null) {
                try {
                    rs.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
        }

        logger.debug("End getDelegation for delegationId=" + delegationId + " dn=" + dn + " localUser=" + localUser);

        return delegation;
    }

    public List<Delegation> getExpiredDelegation() throws DelegationException {
        logger.debug("Begin getExpiredDelegation");

        Connection connection = getConnection();

        List<Delegation> result = new ArrayList<Delegation>(0);
        PreparedStatement selectPreparedStatement = null;
        ResultSet rs = null;
        
        try {
            final String sql = "select dlg_id, dn, fqan, vo, local_user, local_user_group, proxy, info, voms_attrs, termination_time, last_update_time, start_time FROM t_credential where termination_time < ?";

            selectPreparedStatement = connection.prepareStatement(sql);
            selectPreparedStatement.setTimestamp(1, new Timestamp(Calendar.getInstance().getTimeInMillis()));

            // execute query, and return number of rows created
            rs = selectPreparedStatement.executeQuery();

            while (rs.next()) {
                result.add(makeDelegation(rs));
            }
        } catch (SQLException sqle) {
            logger.error("Failure on db interaction", sqle);
            throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
        } finally {
            if (selectPreparedStatement != null) {
                try {
                    selectPreparedStatement.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1);
                }
            }

            if (rs != null) {
                try {
                    rs.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
        }
        logger.debug("End getExpiredDelegation");
        return result;
    }

    public void insertDelegation(Delegation delegation) throws DelegationException {
        if (delegation == null) {
            throw new DelegationException("delegation not specified!");
        }

        logger.debug("Begin insertDelegationProxy for delegationId=" + delegation.getId() + " dn=" + delegation.getDN() + " localUser=" + delegation.getLocalUser());

        if (delegation.getLocalUser() == null) {
            delegation.setLocalUser(CEUtils.getLocalUser());
        }

        if (delegation.getLocalUserGroup() == null) {
            delegation.setLocalUserGroup(CEUtils.getLocalUserGroup());
        }

        Connection connection = getConnection();

        final String sql = "INSERT INTO t_credential (dlg_id, dn, fqan, vo, local_user, local_user_group, proxy, info, voms_attrs, termination_time, last_update_time, start_time) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
        PreparedStatement insertStatement = null;

        try {
            String vomsAttrsStr = "";
            for (String vomsAttribute : delegation.getVOMSAttributes()) {
                vomsAttrsStr += "\t" + vomsAttribute;
            }

            insertStatement = connection.prepareStatement(sql);
            insertStatement.setString(1, delegation.getId());
            insertStatement.setString(2, DNHandler.getDN(delegation.getDN()).getX500());
            insertStatement.setString(3, delegation.getFQAN());
            insertStatement.setString(4, delegation.getVO());
            insertStatement.setString(5, delegation.getLocalUser());
            insertStatement.setString(6, delegation.getLocalUserGroup());
            insertStatement.setString(7, delegation.getProxy());
            insertStatement.setString(8, delegation.getInfo());
            insertStatement.setString(9, vomsAttrsStr);

            if (delegation.getExpirationTime() != null) {
                insertStatement.setTimestamp(10, new java.sql.Timestamp(delegation.getExpirationTime().getTime()));
            } else {
                insertStatement.setTimestamp(10, null);
            }

            if (delegation.getLastUpdateTime() != null) {
                insertStatement.setTimestamp(11, new java.sql.Timestamp(delegation.getLastUpdateTime().getTime()));
            } else {
                insertStatement.setTimestamp(11, null);
            }

            if (delegation.getStartTime() != null) {
                insertStatement.setTimestamp(12, new java.sql.Timestamp(delegation.getStartTime().getTime()));
            } else {
                insertStatement.setTimestamp(12, null);
            }

            insertStatement.executeUpdate();

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            logger.error("Failure on db interaction", sqle);
            try {
                connection.rollback();
            } catch (SQLException e) {
                logger.error(e);
            }
            throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
        } finally {
            if (insertStatement != null) {
                try {
                    insertStatement.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
        }

        logger.debug("End insertDelegationProxy for delegationId=" + delegation.getId() + " dn=" + delegation.getDN() + " localUser=" + delegation.getLocalUser());
    }

    public List<Delegation> listDelegations(String dn, String localUser) throws DelegationException {
        logger.debug("Begin listDelegations");

        Connection connection = getConnection();

        List<Delegation> result = new ArrayList<Delegation>(0);
        PreparedStatement selectPreparedStatement = null;
        ResultSet rs = null;
        
        try {
            final String sql = "select dlg_id, dn, fqan, vo, local_user, local_user_group, proxy, info, voms_attrs, termination_time, last_update_time, start_time FROM t_credential WHERE dn = ? AND local_user = ?";
            selectPreparedStatement = connection.prepareStatement(sql);
            selectPreparedStatement.setString(1, DNHandler.getDN(dn).getX500());
            selectPreparedStatement.setString(2, localUser);
            
            // execute query, and return number of rows created
            rs = selectPreparedStatement.executeQuery();

            while (rs.next()) {
                result.add(makeDelegation(rs));
            }
        } catch (SQLException sqle) {
            logger.error("Failure on db interaction", sqle);
            throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
        } finally {
            if (selectPreparedStatement != null) {
                try {
                    selectPreparedStatement.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1);
                }
            }
            
            if (rs != null) {
                try {
                    rs.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
            
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
        }
        logger.debug("End listDelegations");
        return result;
    }

    public void updateDelegation(Delegation delegation) throws DelegationException {
        if (delegation == null) {
            throw new DelegationException("delegation not specified!");
        }

        if (delegation.getLocalUser() == null) {
            delegation.setLocalUser(CEUtils.getLocalUser());
        }

        if (delegation.getLocalUserGroup() == null) {
            delegation.setLocalUserGroup(CEUtils.getLocalUserGroup());
        }

        logger.debug("Begin updateDelegationProxy for delegationId=" + delegation.getId() + " dn=" + delegation.getDN() + " localUser=" + delegation.getLocalUser());

        delegation.setLastUpdateTime(Calendar.getInstance().getTime());

        Connection connection = getConnection();

        String sql = "UPDATE t_credential C SET proxy = ?, info = ?, voms_attrs = ?, termination_time = ?, last_update_time = ?, start_time = ? WHERE dlg_id = ? AND dn = ?";
        PreparedStatement updateStatement = null;

        try {
            String vomsAttrsStr = "";
            for (String vomsAttribute : delegation.getVOMSAttributes()) {
                vomsAttrsStr += "\t" + vomsAttribute;
            }

            updateStatement = connection.prepareStatement(sql);
            updateStatement.setString(1, delegation.getProxy());
            updateStatement.setString(2, delegation.getInfo());
            updateStatement.setString(3, vomsAttrsStr);
            updateStatement.setTimestamp(4, new java.sql.Timestamp(delegation.getExpirationTime().getTime()));
            updateStatement.setTimestamp(5, new java.sql.Timestamp(delegation.getLastUpdateTime().getTime()));
            updateStatement.setTimestamp(6, new java.sql.Timestamp(delegation.getStartTime().getTime()));
            updateStatement.setString(7, delegation.getId());
            updateStatement.setString(8, DNHandler.getDN(delegation.getDN()).getX500());
            updateStatement.executeUpdate();

            // Commit
            connection.commit();
        } catch (SQLException sqle) {
            logger.error("Failure on db interaction", sqle);
            try {
                connection.rollback();
            } catch (SQLException e) {
                logger.error(e);
            }
            throw new DelegationException("Failure on db interaction: " + sqle.getMessage());
        } finally {
            if (updateStatement != null) {
                try {
                    updateStatement.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error(e);
                }
            }
        }

        logger.debug("End updateDelegationProxy for delegationId=" + delegation.getId() + " dn=" + delegation.getDN() + " localUser=" + delegation.getLocalUser());
    }
}

