/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Paolo Andreetto <paolo.andreetto@pd.infn.it>
 *
 */


package org.glite.ce.cream.ws.delegation;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.glite.ce.cream.configuration.ServiceConfig;
import org.glite.ce.cream.configuration.CREAMConfigConstants.ServiceAttribute;
import org.glite.ce.cream.ws.delegation.storage.CREAMGrDPStorageInterface;
import org.glite.security.SecurityContext;
import org.glite.security.delegation.Delegation;
import org.glite.security.delegation.DelegationException;
import org.glite.security.delegation.GrDPX509Util;
import org.glite.security.delegation.GrDProxyDlgeeOptions;
import org.glite.security.delegation.NewProxyReq;
import org.glite.security.delegation.impl.GliteDelegation;
import org.glite.security.delegation.storage.GrDPStorage;
import org.glite.security.delegation.storage.GrDPStorageElement;
import org.glite.security.delegation.storage.GrDPStorageException;
import org.glite.security.util.axis.InitSecurityContext;

public class CREAMDelegationService implements Delegation {
    private static final Logger logger = Logger.getLogger(CREAMDelegationService.class.getName());

    // Logic implementation class
    private static GliteDelegation gliteDlg = null;

    private static GrDPStorage storageTemporary = null;

    /**
     * The name of the version file for the services.
     */
    protected static final String VERSION_PROPERTIES_FILE = "delegation.version.properties";

    /**
     * The singleton object.
     */
    protected static Properties props = null;

    /**
     * Returns the value of the given property, or <code>null</code> if there is
     * no such property.
     */
    protected static String getProperty(String key) {
        String value = props.getProperty(key);
        logger.debug("Property " + key + " = " + value);

        return value;
    }

    protected static void InitProperties() {
        // return if the singleton is already initialized
        if (props != null) {
            return;
        }

        // singleton initialization
        props = new Properties();

        try {
            InputStream inp = null;

            ClassLoader loader = CREAMDelegationService.class.getClassLoader();

            if (loader != null) {
                inp = loader.getResourceAsStream(VERSION_PROPERTIES_FILE);
            } else {
                inp = ClassLoader.getSystemResourceAsStream(VERSION_PROPERTIES_FILE);
            }

            props.load(new BufferedInputStream(inp));
            inp.close();

            logger.info("Configuration file '" + VERSION_PROPERTIES_FILE + "' loaded");
        } catch (IOException e) {
            e.printStackTrace();
            logger.error("Error loading config file " + VERSION_PROPERTIES_FILE + ": " + e);
        }
    }

    /**
     * Initializes the class implementing the delegation logic.
     */
    public CREAMDelegationService() throws GrDPStorageException {
        if (gliteDlg == null) {
            ServiceConfig serviceConf = ServiceConfig.getConfiguration();

            if (serviceConf == null) {
                throw new GrDPStorageException("Cannot configure delegation service");
            }

            if (serviceConf.getStringParameter(ServiceAttribute.DELEGATION_STORAGE_LABEL) == null) {
                throw new GrDPStorageException("Configuration error: " + ServiceAttribute.DELEGATION_STORAGE_LABEL + " property not defined!");
            }

            if (serviceConf.getStringParameter(ServiceAttribute.DELEGATION_FACTORY_LABEL) == null) {
                throw new GrDPStorageException("Configuration error: " + ServiceAttribute.DELEGATION_FACTORY_LABEL + " property not defined!");
            }
            
            // for version and service metadata
            InitProperties();

            // Sanity check: see if the module version is known
            if (getProperty("module.version") == null) {
                logger.fatal("*** FATAL CONFIGURATION ERROR *** The module.version is unknown ***");
            }

            GrDProxyDlgeeOptions dlgeeOpt = new GrDProxyDlgeeOptions();
            dlgeeOpt.setDlgeeStorage(serviceConf.getStringParameter(ServiceAttribute.DELEGATION_STORAGE_LABEL));
            dlgeeOpt.setDlgeeStorageFactory(serviceConf.getStringParameter(ServiceAttribute.DELEGATION_FACTORY_LABEL));
            dlgeeOpt.setDlgeeStorageDbPool(serviceConf.getStringParameter(ServiceAttribute.DELEGATION_STORAGE_DB_LABEL));
            dlgeeOpt.setDlgeeKeySize(serviceConf.getIntParameter(ServiceAttribute.DELEGATION_KEYSIZE_LABEL, 1024));

            if (storageTemporary == null) {
                try {
                    storageTemporary = GrDPX509Util.getGrDPStorageFactory(dlgeeOpt.getDlgeeStorageFactory()).createGrDPStorage(dlgeeOpt);                    

                } catch (Throwable e) {
                    logger.error("Failed to get a GrDPStorage instance. Delegation is not active." + e.getMessage(), e);
                    throw new GrDPStorageException("Configuration error: " + ServiceAttribute.DELEGATION_FACTORY_LABEL + e.getMessage());
                }
            }

            if (gliteDlg == null) {
                try {
                    gliteDlg = new GliteDelegation(dlgeeOpt);
                } catch (Throwable e) {
                    logger.error("Failed to get a GliteDelegation instance. Delegation is not active." + e.getMessage(), e);
                    throw new GrDPStorageException("Configuration error: failed to get a GliteDelegation instance. Delegation is not active." + e.getMessage());
                }
            }            
        }
    }

    public GrDPStorage getGrDPStorage() {
        return storageTemporary;
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#destroy(java.lang.String)
     */
    public void destroy(String dlgID) throws RemoteException, DelegationException {
        try {
            gliteDlg.destroy(dlgID);
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }

        if (storageTemporary instanceof CREAMGrDPStorageInterface) {
            ((CREAMGrDPStorageInterface) storageTemporary).destroy();
        }
    }

    public String getInterfaceVersion() throws RemoteException, DelegationException {
        try {
            return getProperty("interface.version");
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getNewProxyReq()
     */
    public NewProxyReq getNewProxyReq() throws RemoteException, DelegationException {
        try {
            org.glite.security.delegation.NewProxyReq implProxyReq = gliteDlg.getNewProxyReq(null);
            NewProxyReq proxyReq = new NewProxyReq();
            proxyReq.setDelegationID(implProxyReq.getDelegationID());
            proxyReq.setProxyRequest(implProxyReq.getProxyRequest());

            return proxyReq;
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getProxyReq(java.lang.String)
     */
    public String getProxyReq(String proxy) throws RemoteException, DelegationException {
        org.glite.security.delegation.NewProxyReq proxyReq = null;

        try {
            proxyReq = gliteDlg.getNewProxyReq(proxy);
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }

        if (proxyReq != null) {
            return proxyReq.getProxyRequest();
        }

        return null;
    }

    public String getServiceMetadata(String key) throws RemoteException, DelegationException {
        try {
            return getProperty(key);
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getTerminationTime(java.lang.String)
     */
    public Calendar getTerminationTime(String dlgID) throws RemoteException, DelegationException {
        try {
            return gliteDlg.getTerminationTime(dlgID);
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    public String getVersion() throws RemoteException, DelegationException {
        try {
            return getProperty("module.version");
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#putProxy(java.lang.String,
     *      java.lang.String)
     */
    public void putProxy(String delegationID, String proxy) throws RemoteException, DelegationException {
        try {
            gliteDlg.putProxy(delegationID, proxy);
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#renewProxyReq(java.lang.String)
     */
    public String renewProxyReq(String dlgID) throws RemoteException, DelegationException {
        try {
            logger.debug("Processing renewProxyReq in CREAMDelegationService");

            GrDPStorageElement elem = null;

            // Init Security Context
            InitSecurityContext.init();

            // Get security context
            SecurityContext sc = SecurityContext.getCurrentContext();
            if (sc == null) {
                logger.error("Failed to get SecurityContext.");
                throw new DelegationException("Failed to get client security information.");
            }
            logger.debug("The SecurityContext is succesfully created");

            // Get client DN
            String clientDN = sc.getClientDN().getRFCDNv2();
            if (clientDN == null) {
                logger.error("Failed to get client DN.");
                throw new DelegationException("Failed to get client DN.");
            }
            logger.debug("Got proxy delegation request from client '" + clientDN + "'");

            try {
                elem = storageTemporary.findGrDPStorageElement(dlgID, clientDN);
                logger.debug("The Delegation ID was found in the storage? " + elem);
            } catch (GrDPStorageException e) {
                logger.error("Failure on storage interaction.", e);
                throw new DelegationException("Internal failure.");
            }

            // Check that the DLG ID had a corresponding delegated credential
            if (elem == null) {
                logger.error("Failed to renew credential as there was no DLG ID '" + dlgID + "' for client '" + clientDN + "'");
                throw new DelegationException("The delegation ID was not delegated. Set a correct Delegation ID");
            } else {
                return gliteDlg.renewProxyReq(dlgID);
            }
        } catch (DelegationException de) {
            logger.error(de.getMsg());
            throw de;
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error: " + rt.getMessage());
        }
    }
}
