/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *          Paolo Andreetto <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.cream.ws.delegation.storage;

import java.util.ArrayList;
import java.util.List;

import org.glite.ce.cream.ws.delegation.DelegationProxyEvent;
import org.glite.ce.cream.ws.delegation.DelegationProxyInfo;
import org.glite.ce.cream.ws.delegation.DelegationProxyListener;
import org.glite.security.delegation.GrDProxyDlgeeOptions;
import org.glite.security.delegation.storage.GrDPStorageElement;
import org.glite.security.delegation.storage.GrDPStorageException;
import org.glite.security.delegation.storage.GrDPStorageFilesystem;

public class CREAMGrDPStorage extends GrDPStorageFilesystem implements CREAMGrDPStorageInterface {
    private static ArrayList<DelegationProxyListener> listeners;
    
    public CREAMGrDPStorage(GrDProxyDlgeeOptions dlgeeOpt) throws GrDPStorageException {
        super(dlgeeOpt);

        if (listeners == null) {
            listeners = new ArrayList<DelegationProxyListener>(0);
        }
    }

    public void addDelegationProxyListener(DelegationProxyListener l) {
        if (l != null) {
            listeners.add(l);
        }
    }

    public void removeDelegationProxyListener(DelegationProxyListener l) {
        if (l != null) {
            listeners.remove(l);
        }
    }

    public DelegationProxyListener[] getDelegationProxyListeners() {
        // return eventDispatcher.getDelegationProxyListeners();
        DelegationProxyListener[] list = new DelegationProxyListener[listeners.size()];
        return (DelegationProxyListener[]) listeners.toArray(list);
    }

    /**
     * Insert new delegation request into storage cache area.
     * 
     * @param elem
     *            Object containing the information about the delegation
     *            request.
     * @throws GrDPStorageException
     *             Failed to store new delegation request in storage cache area.
     */
    public void insertGrDPStorageElement(GrDPStorageElement elem) throws GrDPStorageException {
        if (elem == null) {
            return;
        }

        super.insertGrDPStorageElement(elem);

        DelegationProxyInfo delegProxyInfo = DelegationProxyInfo.getInstance(elem.getDelegationID(), elem.getCertificate());
        fireDelegationProxyEvent(new DelegationProxyEvent(DelegationProxyEvent.DELEGATIONPROXY_ADDED, delegProxyInfo));
    }

    public void updateGrDPStorageElement(GrDPStorageElement elem) throws GrDPStorageException {
        if (elem == null) {
            return;
        }

        super.updateGrDPStorageElement(elem);

        DelegationProxyInfo delegProxyInfo = DelegationProxyInfo.getInstance(elem.getDelegationID(), elem.getCertificate());
        fireDelegationProxyEvent(new DelegationProxyEvent(DelegationProxyEvent.DELEGATIONPROXY_UPDATED, delegProxyInfo));
    }

    private void fireDelegationProxyEvent(DelegationProxyEvent event) {
        for (int j = 0; j < listeners.size(); j++) {
            Object listener = listeners.get(j);

            if (listener instanceof DelegationProxyListener) {
                DelegationProxyListener dpListener = (DelegationProxyListener) listener;

                switch(event.getEventType()) {
                    case DelegationProxyEvent.DELEGATIONPROXY_ADDED:
                        dpListener.delegationProxyAdded(event);
                        break;
                    case DelegationProxyEvent.DELEGATIONPROXY_REMOVED:
                        dpListener.delegationProxyRemoved(event);
                        break;
                    case DelegationProxyEvent.DELEGATIONPROXY_UPDATED:   
                        dpListener.delegationProxyUpdated(event);
                        break;
                }
            }
        }
    }

    public void destroy() {
    }

    public List<GrDPStorageElement> findExpiredGrDPStorageElement() throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");
    }

    public void deleteDelegationProxy(String delegationId, String dn, String localUser) throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");        
    }

    public List<DelegationProxyInfo> findExpiredDelegationProxy() throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");
    }

    public DelegationProxyInfo getDelegationProxy(String delegationId, String dn, String localUser) throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");
    }

    public void insertDelegationProxy(DelegationProxyInfo proxy) throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");
    }

    public void updateDelegationProxy(DelegationProxyInfo proxy) throws GrDPStorageException {
        throw new GrDPStorageException("method not implemented!");
    }
}
