/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.cream.ws.delegation.storage;

import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;
import org.glite.ce.cream.ws.delegation.DelegationProxyEvent;
import org.glite.ce.cream.ws.delegation.DelegationProxyListener;

public class EventDispatcher extends Thread {
    private final static Logger logger = Logger.getLogger(EventDispatcher.class);
    private static EventDispatcher eventDispatcher = null;
    private ArrayList<DelegationProxyListener> listeners;
    private BlockingQueue<DelegationProxyEvent> queue;

    private EventDispatcher() {
        queue = new LinkedBlockingQueue<DelegationProxyEvent>();
        listeners = new ArrayList<DelegationProxyListener>(0); 
        start();
    }

    public static EventDispatcher getInstance() {
        if(eventDispatcher == null) {
            eventDispatcher = new EventDispatcher();
        }
        return eventDispatcher;
    }

    public synchronized void addDelegationProxyListener(DelegationProxyListener l) {
        if (l != null) {
            listeners.add(l);
        }
    }

    public synchronized void removeDelegationProxyListener(DelegationProxyListener l) {
        if (l != null) {
            listeners.remove(l);
        }
    }

    public DelegationProxyListener[] getDelegationProxyListeners() {
        DelegationProxyListener[] list = new DelegationProxyListener[listeners.size()];
        return (DelegationProxyListener[]) listeners.toArray(list);
    }

    public void fireDelegationProxyEvent(DelegationProxyEvent event) throws Exception, IllegalArgumentException {
        logger.debug("Begin fireDelegationProxyEvent");
        if(event == null) {
            logger.error("event not specified");
            throw new IllegalArgumentException("event not specified");
        }
        
        try {
            logger.debug("Calling queue.put for eventType = " + event.getEventType());
            queue.put(event);
            logger.debug("queue.put for eventType = " + event.getEventType() + ": done!");
        } catch (InterruptedException e) {
            throw new Exception(e.getMessage());
        }
        logger.debug("End fireDelegationProxyEvent");
    }
        
    public void run() {
        while(!isInterrupted()) {
            DelegationProxyEvent event;
            try {
                logger.debug("Calling queue.take()");
                event = queue.take();
                logger.debug("queue.take(): done!");
                
                for (DelegationProxyListener dpListener : listeners) {
                    switch (event.getEventType()) {
                    case DelegationProxyEvent.DELEGATIONPROXY_ADDED:
                        logger.debug("Event for DELEGATIONPROXY_ADDED ... calling dpListener.delegationProxyAdded");
                        dpListener.delegationProxyAdded(event);
                        break;
                    case DelegationProxyEvent.DELEGATIONPROXY_REMOVED:
                        logger.debug("Event for DELEGATIONPROXY_REMOVED ... calling delegationProxyRemoved");
                        dpListener.delegationProxyRemoved(event);
                        break;
                    case DelegationProxyEvent.DELEGATIONPROXY_UPDATED:
                        logger.debug("Event for DELEGATIONPROXY_UPDATED ... calling delegationProxyUpdated");
                        dpListener.delegationProxyUpdated(event);
                        break;
                    }
                }
            } catch (InterruptedException e) {
                logger.warn("InterruptedException caught: " + e.getMessage());
            } catch (Throwable t) {
                logger.error("The exception Throwable is " + t.getMessage());
            }            
        }
    }
}
